# @file
# Helper lib to read and parse bitmap graphics files
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import logging
import struct


class BmpColorMap(object):
    STATIC_SIZE = 4
# typedef struct {
#  UINT8   Blue;
#  UINT8   Green;
#  UINT8   Red;
#  UINT8   Reserved;
# } BMP_COLOR_MAP;

    def __init__(self, filestream=None):
        if filestream is None:
            self.Blue = 0
            self.Green = 0
            self.Red = 0
            self.Reserved = 0
        else:
            self.PopulateFromFileStream(filestream)

    #
    # Method to un-serialize from a filestream
    #
    def PopulateFromFileStream(self, fs):
        if fs is None:
            raise Exception("Invalid File stream")

        # only populate from file stream those parts that are complete in the file stream
        offset = fs.tell()
        fs.seek(0, 2)
        end = fs.tell()
        fs.seek(offset)

        if (end - offset) < BmpColorMap.STATIC_SIZE:  # size of the bmp color map
            raise Exception("Invalid file stream size.  %d < Color map Size" % (end - offset))

        # read the Bmp Color Map
        self.Blue = struct.unpack("=B", fs.read(1))[0]
        self.Green = struct.unpack("=B", fs.read(1))[0]
        self.Red = struct.unpack("=B", fs.read(1))[0]
        self.Reserved = struct.unpack("=B", fs.read(1))[0]

    def Print(self):
        logger = logging.get(__name__)
        logger.info("BMP Color Map")
        logger.info("  Blue:           0x%X" % self.Blue)
        logger.info("  Green:          0x%X" % self.Green)
        logger.info("  Red:            0x%X" % self.Red)
        logger.info("  Reserved:       0x%X" % self.Reserved)

    def Write(self, fs):
        fs.write(struct.pack("=B", self.Blue))
        fs.write(struct.pack("=B", self.Green))
        fs.write(struct.pack("=B", self.Red))
        fs.write(struct.pack("=B", self.Reserved))


class BmpObject(object):
    STATIC_FILE_HEADER_SIZE = 14
    STATIC_IMAGE_HEADER_SIZE = 40

# typedef struct {
#  CHAR8         CharB;  < -- Start of FileHeader
#  CHAR8         CharM;
#  UINT32        Size;
#  UINT16        Reserved[2];
#  UINT32        ImageOffset;  <-- Start of pixel data relative to start of FileHeader
#  UINT32        HeaderSize;  < -- Start of BmpHeader
#  UINT32        PixelWidth;
#  UINT32        PixelHeight;
#  UINT16        Planes;          ///< Must be 1
#  UINT16        BitPerPixel;     ///< 1, 4, 8, or 24
#  UINT32        CompressionType;
#  UINT32        ImageSize;       ///< Compressed image size in bytes
#  UINT32        XPixelsPerMeter;
#  UINT32        YPixelsPerMeter;
#  UINT32        NumberOfColors;
#  UINT32        ImportantColors;
# } BMP_IMAGE_HEADER;

    def __init__(self, filestream=None):
        self.logger = logging.getLogger(__name__)
        if filestream is None:
            self.CharB = 'B'
            self.CharM = 'M'
            self.Size = BmpObject.STATIC_STRUCT_SIZE
            self.Rsvd16_1 = 0
            self.Rsvd16_2 = 0
            self.ImageOffset = BmpObject.STATIC_STRUCT_SIZE
            self.HeaderSize = BmpObject.STATIC_STRUCT_SIZE
            self.PixelWidth = 0
            self.PixelHeight = 0
            self.Planes = 1
            self.BitPerPixel = 0
            self.CompressionType = 0
            self.ImageSize = 0
            self.XPixelsPerMeter = 0
            self.YPixelsPerMeter = 0
            self.NumberOfColors = 0
            self.ImportantColors = 0

            self.ImageData = None
            self._Padding = None
            self._PaddingLength = 0
            self.ColorMapList = []
        else:
            self.ImageData = None
            self.Padding = None
            self._PaddingLength = 0
            self.ColorMapList = []
            self.PopulateFromFileStream(filestream)

    def ExpectedColorMapEntires(self):
        if(self.BitPerPixel == 1):
            return 2
        elif(self.BitPerPixel == 4):
            return 16
        elif(self.BitPerPixel == 8):
            return 256
        else:
            return 0
    #
    # Method to un-serialize from a filestream
    #

    def PopulateFromFileStream(self, fs):
        if fs is None:
            raise Exception("Invalid File stream")

        # only populate from file stream those parts that are complete in the file stream
        offset = fs.tell()
        fs.seek(0, 2)
        end = fs.tell()
        fs.seek(offset)

        self.logger.debug("Bmp File size as determined by file is: 0x%X (%d)" % (
            end - offset, end - offset))

        if((end - offset) < BmpObject.STATIC_FILE_HEADER_SIZE):  # size of the static file header data
            raise Exception(
                "Invalid file stream size.  %d < File Header Size" % (end - offset))

        # read the BMP File header
        self.CharB = struct.unpack("=c", fs.read(1))[0]
        self.CharM = struct.unpack("=c", fs.read(1))[0]
        self.Size = struct.unpack("=I", fs.read(4))[0]
        self.Rsvd16_1 = struct.unpack("=H", fs.read(2))[0]
        self.Rsvd16_2 = struct.unpack("=H", fs.read(2))[0]
        self.ImageOffset = struct.unpack("=I", fs.read(4))[0]

        if((end - fs.tell()) < BmpObject.STATIC_IMAGE_HEADER_SIZE):
            raise Exception(
                "Invalid file stream size.  %d < Img Header Size" % (end - fs.tell()))

        # read the BMP Image Header
        self.HeaderSize = struct.unpack("=I", fs.read(4))[0]
        self.PixelWidth = struct.unpack("=I", fs.read(4))[0]
        self.PixelHeight = struct.unpack("=I", fs.read(4))[0]
        self.Planes = struct.unpack("=H", fs.read(2))[0]
        self.BitPerPixel = struct.unpack("=H", fs.read(2))[0]
        self.CompressionType = struct.unpack("=I", fs.read(4))[0]
        self.ImageSize = struct.unpack("=I", fs.read(4))[0]
        self.XPixelsPerMeter = struct.unpack("=I", fs.read(4))[0]
        self.YPixelsPerMeter = struct.unpack("=I", fs.read(4))[0]
        self.NumberOfColors = struct.unpack("=I", fs.read(4))[0]
        self.ImportantColors = struct.unpack("=I", fs.read(4))[0]

        if(self.Size < self.HeaderSize):
            raise Exception("Size can't be smaller than HeaderSize")

        if((end - fs.tell()) < (self.Size - self.HeaderSize - BmpObject.STATIC_FILE_HEADER_SIZE)):
            raise Exception("Invalid file stream size (Size) 0x%X Less Than 0x%X" % (
                (end - fs.tell()), (self.Size - self.HeaderSize - BmpObject.STATIC_FILE_HEADER_SIZE)))

        StartOfImageData = offset + self.ImageOffset
        if(fs.tell() < StartOfImageData):

            # Handle any color maps
            if(self.ExpectedColorMapEntires() > 0):
                ColorMapCount = self.ExpectedColorMapEntires()
                if(self.NumberOfColors > 0) and (self.NumberOfColors != ColorMapCount):
                    self.logger.info("Current Code has untested support for limited color map, Good Luck. ")
                    self.logger.info("Expected Color Map Entries %d" % (ColorMapCount))
                    self.logger.info("Actual Color Map Entries %d" % (self.NumberOfColors))
                    ColorMapCount = self.NumberOfColors

                if((StartOfImageData - fs.tell()) < (ColorMapCount * BmpColorMap.STATIC_SIZE)):
                    raise Exception("Color Map not as expected")

                # read all the color maps and append to the list
                for i in range(ColorMapCount):
                    self.ColorMapList.append(BmpColorMap(fs))

            # handle padding
            self._PaddingLength = StartOfImageData - fs.tell()
            self._Padding = fs.read(self._PaddingLength)

        self.ImageData = fs.read(self.Size - self.ImageOffset)

        if((end - fs.tell()) > 0):
            raise Exception(
                "Extra Data at the end of BMP file - 0x%X bytes" % (end - fs.tell()))

    #
    # Method to Print Bmp Header to stdout
    #

    def Print(self, PrintImageData=False, PrintColorMapData=False):
        self.logger.info("BMP")
        self.logger.info("  BMP File Header")
        self.logger.info("    CharB:           %s" % self.CharB)
        self.logger.info("    CharM:           %s" % self.CharM)
        self.logger.info("    Size:            0x%X (%d bytes)" % (self.Size, self.Size))
        self.logger.info("    RSVD[1]:         0x%X" % self.Rsvd16_1)
        self.logger.info("    RSVD[2]:         0x%X" % self.Rsvd16_2)
        self.logger.info("    ImageOffset:     0x%X (%d)" % (self.ImageOffset, self.ImageOffset))
        self.logger.info("  BMP Image Header")
        self.logger.info("    HeaderSize:      0x%X" % self.HeaderSize)
        self.logger.info("    PixelWidth:      0x%X (%d)" % (self.PixelWidth, self.PixelWidth))
        self.logger.info("    PixelHeight:     0x%X (%d)" % (self.PixelHeight, self.PixelHeight))
        self.logger.info("    Planes:          0x%X" % self.Planes)
        self.logger.info("    BitPerPixel:     %d" % self.BitPerPixel)
        self.logger.info("    CompressionType: 0x%X" % self.CompressionType)
        self.logger.info("    ImageSize:       0x%X (used for compressed images only)" % self.ImageSize)
        self.logger.info("    XPixelsPerMeter: %d" % self.XPixelsPerMeter)
        self.logger.info("    YPixelsPerMeter: %d" % self.YPixelsPerMeter)
        self.logger.info("    NumberOfColors:  %d" % self.NumberOfColors)
        self.logger.info("    ImportantColors: %d" % self.ImportantColors)
        # print color maps
        if(PrintColorMapData):
            for cm in self.ColorMapList:
                cm.Print()

        if(self._PaddingLength > 0):
            self.logger.info("  BMP Padding (0x%X bytes)" % self._PaddingLength)
            ndbl = memoryview(self._Padding).tolist()
            for index in range(len(ndbl)):
                if(index % 16 == 0):
                    self.logger.info("0x%04X -" % index),
                self.logger.info(" %02X" % ndbl[index]),
                if(index % 16 == 15):
                    self.logger.info("")
            self.logger.info("")

        if self.ImageData is not None and (PrintImageData):
            self.logger.info("  Bmp Image Data:    ")
            ndbl = memoryview(self.ImageData).tolist()
            for index in range(len(ndbl)):
                if(index % 16 == 0):
                    self.logger.info("0x%04X -" % index),
                self.logger.info(" %02X" % ndbl[index]),
                if(index % 16 == 15):
                    self.logger.info("")
            self.logger.info("")

    def Write(self, fs):
        # Bmp File header
        fs.write(struct.pack("=c", self.CharB))
        fs.write(struct.pack("=c", self.CharM))
        fs.write(struct.pack("=I", self.Size))
        fs.write(struct.pack("=H", self.Rsvd16_1))
        fs.write(struct.pack("=H", self.Rsvd16_2))
        fs.write(struct.pack("=I", self.ImageOffset))

        # Bmp Img Header
        fs.write(struct.pack("=I", self.HeaderSize))
        fs.write(struct.pack("=I", self.PixelWidth))
        fs.write(struct.pack("=I", self.PixelHeight))
        fs.write(struct.pack("=H", self.Planes))
        fs.write(struct.pack("=H", self.BitPerPixel))
        fs.write(struct.pack("=I", self.CompressionType))
        fs.write(struct.pack("=I", self.ImageSize))
        fs.write(struct.pack("=I", self.XPixelsPerMeter))
        fs.write(struct.pack("=I", self.YPixelsPerMeter))
        fs.write(struct.pack("=I", self.NumberOfColors))
        fs.write(struct.pack("=I", self.ImportantColors))

        # Bmp Color Map
        for cm in self.ColorMapList:
            cm.Write(fs)

        # padding
        if(self._PaddingLength > 0):
            fs.write(self.Padding)

        # Pixel data
        if(self.ImageData):
            fs.write(self.ImageData)



## @file
# UnitTest for inf_generator.py
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import unittest
from edk2toollib.windows.capsule.inf_generator import InfGenerator


class InfGeneratorTest(unittest.TestCase):
    VALID_GUID_STRING = "3cad7a0c-d35b-4b75-96b1-03a9fb07b7fc"

    def test_valid(self):
        o = InfGenerator("test_name", "provider", InfGeneratorTest.VALID_GUID_STRING, "x64",
                         "description", "aa.bb.cc.dd", "0xaabbccdd")
        self.assertIsInstance(o, InfGenerator)
        self.assertEqual(o.Name, "test_name")
        self.assertEqual(o.Provider, "provider")
        self.assertEqual(o.EsrtGuid, InfGeneratorTest.VALID_GUID_STRING)
        self.assertEqual(o.Arch, InfGenerator.SUPPORTED_ARCH["x64"])
        self.assertEqual(o.Description, "description")
        self.assertEqual(int(o.VersionHex, 0), int("0xaabbccdd", 0))
        self.assertEqual(o.VersionString, "aa.bb.cc.dd")
        self.assertEqual(o.Manufacturer, "provider")

        # loop thru all supported arch and make sure it works
        for a in InfGenerator.SUPPORTED_ARCH.keys():
            with self.subTest(Arch=a):
                o.Arch = a
                self.assertEqual(InfGenerator.SUPPORTED_ARCH[a], o.Arch)

        # set manufacturer
        o.Manufacturer = "manufacturer"
        self.assertEqual("manufacturer", o.Manufacturer)

    def test_invalid_name_symbol(self):

        InvalidChars = ['~', '`', '!', '@', '#', '$', '%', '^', '&', '*', '(', ')', ' ', '{', '[', '}', ']', '+', '=']
        for a in InvalidChars:
            with self.subTest(name="test{}name".format(a)):
                name = "test{}name".format(a)
                with self.assertRaises(ValueError):
                    InfGenerator(name, "provider", InfGeneratorTest.VALID_GUID_STRING, "x64",
                                 "description", "aa.bb", "0xaabbccdd")

    def test_version_string_format(self):
        with self.subTest(version_string="zero ."):
            with self.assertRaises(ValueError):
                InfGenerator("test_name", "provider", InfGeneratorTest.VALID_GUID_STRING, "x64",
                             "description", "1234", "0x100000000")

        with self.subTest(version_string="> 3 ."):
            with self.assertRaises(ValueError):
                InfGenerator("test_name", "provider", InfGeneratorTest.VALID_GUID_STRING, "x64",
                             "description", "1.2.3.4.5", "0x100000000")

    def test_version_hex_too_big(self):
        with self.subTest("hex string too big"):
            with self.assertRaises(ValueError):
                InfGenerator("test_name", "provider", InfGeneratorTest.VALID_GUID_STRING, "x64",
                             "description", "aa.bb", "0x100000000")

        with self.subTest("decimal too big"):
            with self.assertRaises(ValueError):
                InfGenerator("test_name", "provider", InfGeneratorTest.VALID_GUID_STRING, "x64",
                             "description", "aa.bb", "4294967296")

    def test_version_hex_can_support_decimal(self):
        o = InfGenerator("test_name", "provider", InfGeneratorTest.VALID_GUID_STRING, "x64",
                         "description", "aa.bb.cc.dd", "12356")
        self.assertEqual(int(o.VersionHex, 0), 12356)

    def test_invalid_guid_format(self):
        with self.assertRaises(ValueError):
            InfGenerator("test_name", "provider", "NOT A VALID GUID", "x64", "description", "aa.bb", "0x1000000")



# @file inf_parser.py
# Code to help parse EDK2 INF files
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
from edk2toollib.uefi.edk2.parsers.base_parser import HashFileParser
import os


AllPhases = ["SEC", "PEIM", "PEI_CORE", "DXE_DRIVER", "DXE_CORE", "DXE_RUNTIME_DRIVER", "UEFI_DRIVER",
             "SMM_CORE", "DXE_SMM_DRIVER", "UEFI_APPLICATION"]


class InfParser(HashFileParser):

    def __init__(self):
        HashFileParser.__init__(self, 'ModuleInfParser')
        self.Lines = []
        self.Parsed = False
        self.Dict = {}
        self.LibraryClass = ""
        self.SupportedPhases = []
        self.PackagesUsed = []
        self.LibrariesUsed = []
        self.ProtocolsUsed = []
        self.GuidsUsed = []
        self.PpisUsed = []
        self.PcdsUsed = []
        self.Sources = []
        self.Binaries = []
        self.Path = ""

    def ParseFile(self, filepath):
        self.Logger.debug("Parsing file: %s" % filepath)
        if(not os.path.isabs(filepath)):
            fp = self.FindPath(filepath)
        else:
            fp = filepath
        self.Path = fp
        f = open(fp, "r")
        self.Lines = f.readlines()
        f.close()
        InDefinesSection = False
        InPackagesSection = False
        InLibraryClassSection = False
        InProtocolsSection = False
        InGuidsSection = False
        InPpiSection = False
        InPcdSection = False
        InSourcesSection = False
        InBinariesSection = False

        for line in self.Lines:
            sline = self.StripComment(line)

            if(sline is None or len(sline) < 1):
                continue

            if InDefinesSection:
                if sline.strip()[0] == '[':
                    InDefinesSection = False
                else:
                    if sline.count("=") == 1:
                        tokens = sline.split('=', 1)
                        self.Dict[tokens[0].strip()] = tokens[1].strip()
                        #
                        # Parse Library class and phases in special manor
                        #
                        if(tokens[0].strip().lower() == "library_class"):
                            self.LibraryClass = tokens[1].partition("|")[0].strip()
                            self.Logger.debug("Library class found")
                            if(len(tokens[1].partition("|")[2].strip()) < 1):
                                self.SupportedPhases = AllPhases
                            elif(tokens[1].partition("|")[2].strip().lower() == "base"):
                                self.SupportedPhases = AllPhases
                            else:
                                self.SupportedPhases = tokens[1].partition("|")[2].strip().split()

                        self.Logger.debug("Key,values found:  %s = %s" % (tokens[0].strip(), tokens[1].strip()))

                        continue

            elif InPackagesSection:
                if sline.strip()[0] == '[':
                    InPackagesSection = False
                else:
                    self.PackagesUsed.append(sline.partition("|")[0].strip())
                    continue

            elif InLibraryClassSection:
                if sline.strip()[0] == '[':
                    InLibraryClassSection = False
                else:
                    self.LibrariesUsed.append(sline.partition("|")[0].strip())
                    continue

            elif InProtocolsSection:
                if sline.strip()[0] == '[':
                    InProtocolsSection = False
                else:
                    self.ProtocolsUsed.append(sline.partition("|")[0].strip())
                    continue

            elif InGuidsSection:
                if sline.strip()[0] == '[':
                    InGuidsSection = False
                else:
                    self.GuidsUsed.append(sline.partition("|")[0].strip())
                    continue

            elif InPcdSection:
                if sline.strip()[0] == '[':
                    InPcdSection = False
                else:
                    self.PcdsUsed.append(sline.partition("|")[0].strip())
                    continue

            elif InPpiSection:
                if sline.strip()[0] == '[':
                    InPpiSection = False
                else:
                    self.PpisUsed.append(sline.partition("|")[0].strip())
                    continue

            elif InSourcesSection:
                if sline.strip()[0] == '[':
                    InSourcesSection = False
                else:
                    self.Sources.append(sline.partition("|")[0].strip())
                    continue

            elif InBinariesSection:
                if sline.strip()[0] == '[':
                    InBinariesSection = False
                else:
                    self.Binaries.append(sline.partition("|")[0].strip())
                    continue

            # check for different sections
            if sline.strip().lower().startswith('[defines'):
                InDefinesSection = True

            elif sline.strip().lower().startswith('[packages'):
                InPackagesSection = True

            elif sline.strip().lower().startswith('[libraryclasses'):
                InLibraryClassSection = True

            elif sline.strip().lower().startswith('[protocols'):
                InProtocolsSection = True

            elif sline.strip().lower().startswith('[ppis'):
                InPpiSection = True

            elif sline.strip().lower().startswith('[guids'):
                InGuidsSection = True

            elif sline.strip().lower().startswith('[pcd') or \
                    sline.strip().lower().startswith('[patchpcd') or \
                    sline.strip().lower().startswith('[fixedpcd') or \
                    sline.strip().lower().startswith('[featurepcd'):
                InPcdSection = True

            elif sline.strip().lower().startswith('[sources'):
                InSourcesSection = True

            elif sline.strip().lower().startswith('[binaries'):
                InBinariesSection = True

        self.Parsed = True



##
# junit_report_format
# This module contains support for Outputting Junit test results xml.
#
# Used to support CI/CD and exporting test results for other tools.
# This does test report generation without being a test runner.
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import time


class JunitReportError(object):
    def __init__(self, type, msg):
        self.Message = msg
        self.Type = type


class JunitReportFailure(object):
    def __init__(self, type, msg):
        self.Message = msg
        self.Type = type

##
# Test Case class
#
##


class JunitReportTestCase(object):
    NEW = 1
    SKIPPED = 2
    FAILED = 3
    ERROR = 4
    SUCCESS = 5

    def __init__(self, Name, ClassName):
        self.Name = Name
        self.ClassName = ClassName
        self.Time = 0
        self.Status = JunitReportTestCase.NEW

        self.FailureMsg = None
        self.ErrorMsg = None
        self._TestSuite = None
        self.StdErr = ""
        self.StdOut = ""
        self._StartTime = time.time()

    def SetFailed(self, Msg, Type):
        if(self.Status != JunitReportTestCase.NEW):
            raise Exception("Can't Set to failed.  State must be in NEW")
        self.Time = time.time() - self._StartTime
        self.Status = JunitReportTestCase.FAILED
        self.FailureMsg = JunitReportFailure(Type, Msg)

    def SetError(self, Msg, Type):
        if(self.Status != JunitReportTestCase.NEW):
            raise Exception("Can't Set to error.  State must be in NEW")
        self.Time = time.time() - self._StartTime
        self.Status = JunitReportTestCase.ERROR
        self.ErrorMsg = JunitReportError(Type, Msg)

    def SetSuccess(self):
        if(self.Status != JunitReportTestCase.NEW):
            raise Exception("Can't Set to success.  State must be in NEW")
        self.Status = JunitReportTestCase.SUCCESS
        self.Time = time.time() - self._StartTime

    def SetSkipped(self):
        if(self.Status != JunitReportTestCase.NEW):
            raise Exception("Can't Set to skipped.  State must be in NEW")
        self.Status = JunitReportTestCase.SKIPPED
        self.Time = time.time() - self._StartTime

    def LogStdOut(self, msg):
        self.StdOut += msg.strip() + "\n "

    def LogStdError(self, msg):
        self.StdErr += msg.strip() + "\n "

    def Output(self, outstream):
        outstream.write('<testcase classname="{0}" name="{1}" time="{2}">'.format(self.ClassName, self.Name, self.Time))
        if self.Status == JunitReportTestCase.SKIPPED:
            outstream.write('<skipped />')
        elif self.Status == JunitReportTestCase.FAILED:
            outstream.write('<failure message="{0}" type="{1}" />'.format(self.FailureMsg.Message,
                                                                          self.FailureMsg.Type))
        elif self.Status == JunitReportTestCase.ERROR:
            outstream.write('<error message="{0}" type="{1}" />'.format(self.ErrorMsg.Message, self.ErrorMsg.Type))
        elif self.Status != JunitReportTestCase.SUCCESS:
            raise Exception("Can't output a testcase {0}.{1} in invalid state {2}".format(self.ClassName,
                                                                                          self.Name, self.Status))

        outstream.write('<system-out>' + self.StdOut + '</system-out>')
        outstream.write('<system-err>' + self.StdErr + '</system-err>')
        outstream.write('</testcase>')


##
# Test Suite class.  Create new suites by using the JunitTestReport Object
#
#
##
class JunitReportTestSuite(object):
    def __init__(self, Name, Package, Id):
        self.Name = Name
        self.Package = Package
        self.TestId = Id
        self.TestCases = []

    def create_new_testcase(self, name, classname):
        tc = JunitReportTestCase(name, classname)
        self.TestCases.append(tc)
        tc._TestSuite = self
        return tc

    def Output(self, outstream):
        Errors = 0
        Failures = 0
        Skipped = 0
        Tests = len(self.TestCases)

        for a in self.TestCases:
            if(a.Status == JunitReportTestCase.FAILED):
                Failures += 1
            elif(a.Status == JunitReportTestCase.ERROR):
                Errors += 1
            elif(a.Status == JunitReportTestCase.SKIPPED):
                Skipped += 1

        outstream.write('<testsuite id="{0}" name="{1}" package="{2}" errors="{3}" tests="{4}" '
                        'failures="{5}" skipped="{6}">'.format(self.TestId, self.Name, self.Package,
                                                               Errors, Tests, Failures, Skipped))

        for a in self.TestCases:
            a.Output(outstream)

        outstream.write('</testsuite>')

##
# Test Report.  Top level object test reporting.
#
#
##


class JunitTestReport(object):
    def __init__(self):
        self.TestSuites = []

    def create_new_testsuite(self, name, package):
        id = len(self.TestSuites)
        ts = JunitReportTestSuite(name, package, id)
        self.TestSuites.append(ts)
        return ts

    def Output(self, filepath):
        f = open(filepath, "w")
        f.write('')
        f.write('<?xml version="1.0" encoding="UTF-8"?>')
        f.write('<testsuites>')
        for a in self.TestSuites:
            a.Output(f)
        f.write('</testsuites>')
        f.close()



##
# Handle basic logging outputting to files
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import logging


class FileHandler(logging.FileHandler):
    def __init__(self, filename, mode='w+'):
        logging.FileHandler.__init__(self, filename, mode=mode)

    def handle(self, record):
        """
        Conditionally emit the specified logging record.
        Emission depends on filters which may have been added to the handler.
        Wrap the actual emission of the record with acquisition/release of
        the I/O thread lock. Returns whether the filter passed the record for
        emission.
        """

        rv = self.filter(record)
        if rv and record.levelno >= self.level:
            self.acquire()
            try:
                self.emit(record)
            finally:
                self.release()
        return rv



##
# Module to generate inf files for capsule update based on INF TEMPLATE and
# supplied information (Name, Version, ESRT Guid, Rollback, etc.)
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import os
import logging
import datetime
import re
import uuid


class InfGenerator(object):

    ### INF Template ###
    TEMPLATE = r""";
; {Name}.inf
; {DriverVersion}
; Copyright (C) 2019 Microsoft Corporation.  All Rights Reserved.
;
[Version]
Signature="$WINDOWS NT$"
Class=Firmware
ClassGuid={{f2e7dd72-6468-4e36-b6f1-6488f42c1b52}}
Provider=%Provider%
DriverVer={Date},{DriverVersion}
PnpLockdown=1
CatalogFile={Name}.cat

[Manufacturer]
%MfgName% = Firmware,NT{Arch}

[Firmware.NT{Arch}]
%FirmwareDesc% = Firmware_Install,UEFI\RES_{{{EsrtGuid}}}

[Firmware_Install.NT]
CopyFiles = Firmware_CopyFiles
{Rollback}
[Firmware_CopyFiles]
{FirmwareBinFile}

[Firmware_Install.NT.Hw]
AddReg = Firmware_AddReg

[Firmware_AddReg]
HKR,,FirmwareId,,{{{EsrtGuid}}}
HKR,,FirmwareVersion,%REG_DWORD%,{VersionHexString}
HKR,,FirmwareFilename,,{FirmwareBinFile}

[SourceDisksNames]
1 = %DiskName%

[SourceDisksFiles]
{FirmwareBinFile} = 1

[DestinationDirs]
DefaultDestDir = %DIRID_WINDOWS%,Firmware ; %SystemRoot%\Firmware

[Strings]
; localizable
Provider     = "{Provider}"
MfgName      = "{MfgName}"
FirmwareDesc = "{Description}"
DiskName     = "Firmware Update"

; non-localizable
DIRID_WINDOWS = 10
REG_DWORD     = 0x00010001
"""

    ROLLBACKTEMPLATE = r"""AddReg    = Firmware_DowngradePolicy_Addreg

  ;override firmware resource update policy to allow downgrade to lower version
  [Firmware_DowngradePolicy_Addreg]
  HKLM,SYSTEM\CurrentControlSet\Control\FirmwareResources\{{{EsrtGuid}}},Policy,%REG_DWORD%,1
  """

    SUPPORTED_ARCH = {'amd64': 'amd64',
                      'x64': 'amd64',
                      'arm': 'arm',
                      'arm64': 'ARM64',
                      'aarch64': 'ARM64'
                      }

    def __init__(self, name_string, provider, esrt_guid, arch, description_string, version_string, version_hex):
        self.Name = name_string
        self.Provider = provider
        self.EsrtGuid = esrt_guid
        self.Arch = arch
        self.Description = description_string
        self.VersionString = version_string
        self.VersionHex = version_hex
        self._manufacturer = None  # default for optional feature
        self._date = datetime.date.today()

    @property
    def Name(self):
        return self._name

    @Name.setter
    def Name(self, value):
        # test here for invalid chars
        if not (re.compile(r'[\w-]*$')).match(value):
            logging.critical("Name invalid: '%s'", value)
            raise ValueError("Name has invalid chars.")
        self._name = value

    @property
    def Provider(self):
        return self._provider

    @Provider.setter
    def Provider(self, value):
        self._provider = value

    @property
    def Manufacturer(self):
        if(self._manufacturer is None):
            return self.Provider

        return self._manufacturer

    @Manufacturer.setter
    def Manufacturer(self, value):
        self._manufacturer = value

    @property
    def Description(self):
        return self._description

    @Description.setter
    def Description(self, value):
        self._description = value

    @property
    def EsrtGuid(self):
        return self._esrtguid

    @EsrtGuid.setter
    def EsrtGuid(self, value):
        uuid.UUID(value)  # if this works it is valid...otherwise throws exception
        # todo - make sure it is formatted exactly right
        self._esrtguid = value

    @property
    def VersionString(self):
        return self._versionstring

    @VersionString.setter
    def VersionString(self, value):
        c = value.count(".")
        if(c < 1) or (c > 3):
            logging.critical("Version string in invalid format.")
            raise ValueError("VersionString must be in format of xx.xx -> xx.xx.xx.xx")
        self._versionstring = value

    @property
    def VersionHex(self):
        return "0x%X" % self._versionhex

    @VersionHex.setter
    def VersionHex(self, value):
        a = int(value, 0)
        if(a > 0xFFFFFFFF):
            logging.critical("VersionHex invalid: '%s'", value)
            raise ValueError("VersionHex must fit within 32bit value range for unsigned integer")
        self._versionhex = a

    @property
    def Arch(self):
        return self._arch

    @Arch.setter
    def Arch(self, value):
        key = value.lower()
        if(key not in InfGenerator.SUPPORTED_ARCH.keys()):
            logging.critical("Arch invalid: '%s'", value)
            raise ValueError("Unsupported Architecture")
        self._arch = InfGenerator.SUPPORTED_ARCH[key]

    @property
    def Date(self):
        return self._date.strftime("%m/%d/%Y")

    @Date.setter
    def Date(self, value):
        if(not isinstance(value, datetime.date)):
            raise ValueError("Date must be a datetime.date object")
        self._date = value

    def MakeInf(self, OutputInfFilePath, FirmwareBinFileName, Rollback=False):
        RollbackString = ""
        if(Rollback):
            RollbackString = InfGenerator.ROLLBACKTEMPLATE.format(EsrtGuid=self.EsrtGuid)

        binfilename = os.path.basename(FirmwareBinFileName)

        Content = InfGenerator.TEMPLATE.format(
            Name=self.Name,
            Date=self.Date,
            Arch=self.Arch,
            DriverVersion=self.VersionString,
            EsrtGuid=self.EsrtGuid,
            FirmwareBinFile=binfilename,
            VersionHexString=self.VersionHex,
            Provider=self.Provider,
            MfgName=self.Manufacturer,
            Description=self.Description,
            Rollback=RollbackString)

        with open(OutputInfFilePath, "w") as f:
            f.write(Content)

        return 0



# @file
# Module contains helper classes for working with Fault Tolerant Working block content
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import struct
import uuid
import sys

#
# UEFI GUIDs
#
EdkiiWorkingBlockSignatureGuid = uuid.UUID(fields=(0x9E58292B, 0x7C68, 0x497D, 0xA0, 0xCE, 0x6500FD9F1B95))

#
# The EDKII Fault tolerant working block header.
# The header is immediately followed by the write queue data.
#
# typedef struct {
#   EFI_GUID  Signature;
#   UINT32    Crc;
#   UINT8     WorkingBlockValid : 1;
#   UINT8     WorkingBlockInvalid : 1;
#   UINT8     Reserved : 6;
#   UINT8     Reserved3[3];
#   UINT64    WriteQueueSize;
# } EFI_FAULT_TOLERANT_WORKING_BLOCK_HEADER;


class EfiFtwWorkingBlockHeader(object):
    def __init__(self):
        self.StructString = "=16sLBBBBQ"
        self.Signature = None
        self.Crc = None
        self.WorkingBlockValidFields = None
        self.Reserved1 = None
        self.Reserved2 = None
        self.Reserved3 = None
        self.WriteQueueSize = None

    def load_from_file(self, file):
        # This function assumes that the file has been seeked
        # to the correct starting location.
        orig_seek = file.tell()
        struct_bytes = file.read(struct.calcsize(self.StructString))
        file.seek(orig_seek)

        # Load this object with the contents of the data.
        (signature_bin, self.Crc, self.WorkingBlockValidFields, self.Reserved1, self.Reserved2, self.Reserved3,
         self.WriteQueueSize) = struct.unpack(self.StructString, struct_bytes)

        # Update the GUID to be a UUID object.
        if sys.byteorder == 'big':
            self.Signature = uuid.UUID(bytes=signature_bin)
        else:
            self.Signature = uuid.UUID(bytes_le=signature_bin)

        # Check that signature is valid
        if self.Signature != EdkiiWorkingBlockSignatureGuid:
            raise Exception("FTW Working Block Header has unknown signature: %s" % self.Signature)

        return self

    def serialize(self):
        signature_bin = self.Signature.bytes if sys.byteorder == 'big' else self.Signature.bytes_le
        return struct.pack(self.StructString, signature_bin, self.Crc, self.WorkingBlockValidFields,
                           self.Reserved1, self.Reserved2, self.Reserved3, self.WriteQueueSize)

#
# EFI Fault tolerant block update write queue entry.
#
# typedef struct {
#   UINT8     HeaderAllocated : 1;
#   UINT8     WritesAllocated : 1;
#   UINT8     Complete : 1;
#   UINT8     Reserved : 5;
#   EFI_GUID  CallerId;
#   UINT64    NumberOfWrites;
#   UINT64    PrivateDataSize;
# } EFI_FAULT_TOLERANT_WRITE_HEADER;


class EfiFtwWriteHeader(object):
    def __init__(self):
        self.StructString = "=BBBB16sLQQ"
        self.StatusBits = None
        self.ReservedByte1 = None
        self.ReservedByte2 = None
        self.ReservedByte3 = None
        self.CallerId = None
        self.ReservedUint32 = None
        self.NumberOfWrites = None
        self.PrivateDataSize = None

    def load_from_file(self, file):
        # This function assumes that the file has been seeked
        # to the correct starting location.
        orig_seek = file.tell()
        struct_bytes = file.read(struct.calcsize(self.StructString))
        file.seek(orig_seek)

        # Load this object with the contents of the data.
        (self.StatusBits, self.ReservedByte1, self.ReservedByte2, self.ReservedByte3, self.CallerId,
         self.ReservedUint32, self.NumberOfWrites, self.PrivateDataSize) = struct.unpack(self.StructString,
                                                                                         struct_bytes)

        return self

    def serialize(self):
        return struct.pack(self.StructString, self.StatusBits, self.ReservedByte1, self.ReservedByte2,
                           self.ReservedByte3, self.CallerId, self.ReservedUint32, self.NumberOfWrites,
                           self.PrivateDataSize)

#
# EFI Fault tolerant block update write queue record.
#
# typedef struct {
#   UINT8   BootBlockUpdate : 1;
#   UINT8   SpareComplete : 1;
#   UINT8   DestinationComplete : 1;
#   UINT8   Reserved : 5;
#   EFI_LBA Lba;
#   UINT64  Offset;
#   UINT64  Length;
#   INT64   RelativeOffset;
# } EFI_FAULT_TOLERANT_WRITE_RECORD;


class EfiFtwWriteRecord(object):
    def __init__(self):
        self.StructString = "=BBBBLQQQQ"
        self.StatusBits = None
        self.ReservedByte1 = None
        self.ReservedByte2 = None
        self.ReservedByte3 = None
        self.ReservedUint32 = None
        self.Lba = None
        self.Offset = None
        self.Length = None
        self.RelativeOffset = None

    def load_from_file(self, file):
        # This function assumes that the file has been seeked
        # to the correct starting location.
        orig_seek = file.tell()
        struct_bytes = file.read(struct.calcsize(self.StructString))
        file.seek(orig_seek)

        # Load this object with the contents of the data.
        (self.StatusBits, self.ReservedByte1, self.ReservedByte2, self.ReservedByte3, self.ReservedUint32, self.Lba,
         self.Offset, self.Length, self.RelativeOffset) = struct.unpack(self.StructString, struct_bytes)

        return self

    def serialize(self):
        return struct.pack(self.StructString, self.StatusBits, self.ReservedByte1, self.ReservedByte2,
                           self.ReservedByte3, self.ReservedUint32, self.Lba, self.Offset, self.Length,
                           self.RelativeOffset)



# @file path_utilities.py
# Code to help convert Edk2, absolute, and relative file paths
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import os
import logging
import fnmatch

#
# Class to help convert from absolute path to EDK2 build path
# using workspace and packagepath variables
#


class Edk2Path(object):

    #
    # ws - absolute path or cwd relative to workspace
    # packagepathlist - list of packages path.  Absolute path list or workspace relative path
    #
    def __init__(self, ws, packagepathlist):
        self.WorkspacePath = ws
        self.logger = logging.getLogger("Edk2Path")
        if(not os.path.isabs(ws)):
            self.WorkspacePath = os.path.abspath(os.path.join(os.getcwd(), ws))

        if(not os.path.isdir(self.WorkspacePath)):
            self.logger.error("Workspace path invalid.  {0}".format(ws))
            raise Exception("Workspace path invalid.  {0}".format(ws))

        # Set PackagePath
        self.PackagePathList = list()
        for a in packagepathlist:
            if(os.path.isabs(a)):
                self.PackagePathList.append(a)
            else:
                # see if workspace relative
                wsr = os.path.join(ws, a)
                if(os.path.isdir(wsr)):
                    self.PackagePathList.append(wsr)
                else:
                    # assume current working dir relative.  Will catch invalid dir when checking whole list
                    self.PackagePathList.append(os.path.abspath(os.path.join(os.getcwd(), a)))

        error = False
        for a in self.PackagePathList:
            if(not os.path.isdir(a)):
                self.logger.error("Invalid package path entry {0}".format(a))
                error = True

        # report error
        if(error):
            raise Exception("Invalid package path directory(s)")

    def GetEdk2RelativePathFromAbsolutePath(self, abspath):
        relpath = None
        found = False
        if abspath is None:
            return None
        for a in self.PackagePathList:
            stripped = abspath.lower().partition(a.lower())[2]
            if stripped:
                # found our path...now lets correct for case
                relpath = abspath[len(a):]
                found = True
                self.logger.debug("Successfully converted AbsPath to Edk2Relative Path using PackagePath")
                self.logger.debug("AbsolutePath: %s found in PackagePath: %s" % (abspath, a))
                break

        if(not found):
            # try to strip the workspace
            stripped = abspath.lower().partition(self.WorkspacePath.lower())[2]
            if stripped:
                # found our path...now lets correct for case
                relpath = abspath[len(self.WorkspacePath):]
                found = True
                self.logger.debug("Successfully converted AbsPath to Edk2Relative Path using WorkspacePath")
                self.logger.debug("AbsolutePath: %s found in Workspace: %s" % (abspath, self.WorkspacePath))

        if(found):
            relpath = relpath.replace(os.sep, "/")
            return relpath.lstrip("/")

        # didn't find the path for conversion.
        self.logger.error("Failed to convert AbsPath to Edk2Relative Path")
        self.logger.error("AbsolutePath: %s" % abspath)
        return None

    def GetAbsolutePathOnThisSytemFromEdk2RelativePath(self, relpath):
        if relpath is None:
            return None
        relpath = relpath.replace("/", os.sep)
        abspath = os.path.join(self.WorkspacePath, relpath)
        if os.path.exists(abspath):
            return abspath

        for a in self.PackagePathList:
            abspath = os.path.join(a, relpath)
            if(os.path.exists(abspath)):
                return abspath
        self.logger.error("Failed to convert Edk2Relative Path to an Absolute Path on this system.")
        self.logger.error("Relative Path: %s" % relpath)

        return None

    # Find the package this path belongs to using
    # some Heuristic.  This isn't perfect but at least
    # identifies the directory consistently
    #
    # @param InputPath:  absolute path to module
    #
    # @ret Name of Package that the module is in.
    def GetContainingPackage(self, InputPath):
        self.logger.debug("GetContainingPackage: %s" % InputPath)

        dirpathprevious = os.path.dirname(InputPath)
        dirpath = os.path.dirname(InputPath)
        while(dirpath is not None):
            #
            # if at the root of a packagepath return the previous dir.
            # this catches cases where a package has no DEC
            #
            if(dirpath in self.PackagePathList):
                a = os.path.basename(dirpathprevious)
                self.logger.debug("Reached Package Path.  Using previous directory: %s" % a)
                return a
            #
            # if at the root of the workspace return the previous dir.
            # this catches cases where a package has no DEC
            #
            if(dirpath == self.WorkspacePath):
                a = os.path.basename(dirpathprevious)
                self.logger.debug("Reached Workspace Path.  Using previous directory: %s" % a)
                return a
            #
            # Check for a DEC file in this folder
            # if here then return the directory name as the "package"
            #
            for f in os.listdir(dirpath):
                if fnmatch.fnmatch(f, '*.dec'):
                    a = os.path.basename(dirpath)
                    self.logger.debug("Found DEC file at %s.  Pkg is: %s", dirpath, a)
                    return a

            dirpathprevious = dirpath
            dirpath = os.path.dirname(dirpath)

        self.logger.error("Failed to find containing package for %s" % InputPath)
        self.logger.info("PackagePath is: %s" % os.pathsep.join(self.PackagePathList))
        self.logger.info("Workspace path is : %s" % self.WorkspacePath)
        return None



##
# Handle basic logging outputting to markdown
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import logging


class MarkdownFileHandler(logging.FileHandler):
    def __init__(self, filename, mode='w+'):
        logging.FileHandler.__init__(self, filename, mode=mode)
        if self.stream.writable:
            self.stream.write("  # Build Report\n")
            self.stream.write("[Go to table of contents](#table-of-contents)\n")
            self.stream.write("=====\n")
            self.stream.write(" [Go to Error List](#error-list)\n")
            self.stream.write("=====\n")
        self.contents = []
        self.error_records = []

    def emit(self, record):
        if self.stream is None:
            self.stream = self._open()
        msg = record.message.strip("#- ")

        if len(msg) > 0:
            if logging.getLevelName(record.levelno) == "SECTION":
                self.contents.append((msg, []))
                msg = "## " + msg
            elif record.levelno == logging.CRITICAL:
                section_index = len(self.contents) - 1
                if section_index >= 0:
                    self.contents[section_index][1].append(msg)
                msg = "### " + msg
            elif record.levelno == logging.ERROR:
                self.error_records.append(record)
                msg = "#### ERROR: " + msg
            elif record.levelno == logging.WARNING:
                msg = "  _ WARNING: " + msg + "_"
            else:
                msg = "    " + msg
            stream = self.stream
            # issue 35046: merged two stream.writes into one.
            stream.write(msg + self.terminator)

            # self.flush()

    def handle(self, record):
        """
        Conditionally emit the specified logging record.
        Emission depends on filters which may have been added to the handler.
        Wrap the actual emission of the record with acquisition/release of
        the I/O thread lock. Returns whether the filter passed the record for
        emission.
        """

        rv = self.filter(record)
        if rv and record.levelno >= self.level:
            self.acquire()
            try:
                self.emit(record)
            finally:
                self.release()
        return rv

    @staticmethod
    def __convert_to_markdownlink(text):
        # Using info from here https://stackoverflow.com/a/38507669
        # get rid of uppercase characters
        text = text.lower().strip()
        # get rid of punctuation
        text = text.replace(".", "").replace(",", "").replace("-", "")
        # replace spaces
        text = text.replace(" ", "-")
        return text

    def _output_error(self, record):
        output = " + \"{0}\" from {1}:{2}\n".format(record.msg, record.pathname, record.lineno)
        self.stream.write(output)

    def close(self):
        self.stream.write("## Table of Contents\n")
        for item, subsections in self.contents:
            link = MarkdownFileHandler.__convert_to_markdownlink(item)
            self.stream.write("+ [{0}](#{1})\n".format(item, link))
            for section in subsections:
                section_link = MarkdownFileHandler.__convert_to_markdownlink(section)
                self.stream.write("  + [{0}](#{1})\n".format(section, section_link))

        self.stream.write("## Error List\n")
        if len(self.error_records) == 0:
            self.stream.write("   No errors found")
        for record in self.error_records:
            self._output_error(record)

        self.flush()
        self.stream.close()



# @file buildreport_parser.py
# Code to help parse an EDk2 Build Report
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import os
import logging

import edk2toollib.utility_functions as uf
import edk2toollib.uefi.edk2.path_utilities as pu

#
# Class to represent a module within the Build Report
#


class ModuleSummary(object):
    def __init__(self, content, ws, packagepatahlist):
        self._RawContent = content
        self.Guid = ""
        self.Name = ""
        self.InfPath = ""
        self.Type = ""
        self.PCDs = {}
        self.Libraries = {}
        self.Depex = ""
        self.WorkspacePath = ws
        self.PackagePathList = packagepatahlist
        self.FvName = None

    def Parse(self):
        inPcdSection = False
        inLibSection = False
        inDepSection = False
        nextLineSection = False
        tokenspace = ""
        pathConverter = pu.Edk2Path(self.WorkspacePath, self.PackagePathList)

        i = 0
        try:

            while i < len(self._RawContent):
                line = self._RawContent[i].strip()

                # parse start and end
                if line == ">----------------------------------------------------------------------------------------------------------------------<":      # noqa: E501
                    nextLineSection = True

                elif line == "<---------------------------------------------------------------------------------------------------------------------->":    # noqa: E501
                    inPcdSection = False
                    inLibSection = False
                    inDepSection = False
                    nextLineSection = False

                # parse section header
                elif(nextLineSection):
                    nextLineSection = False
                    if(line == "Library"):
                        inLibSection = True
                        i += 1  # add additional line to skip the dashed line

                    elif(line == "PCD"):
                        inPcdSection = True
                        i += 1  # add additional line to skip the dashed line

                    elif(line == "Final Dependency Expression (DEPEX) Instructions"):
                        inDepSection = True
                        i += 1  # add additional line to skip the dashed line
                    else:
                        logging.debug("Unsupported Section: " + line)
                        inPcdSection = False
                        inLibSection = False
                        inDepSection = False

                # Normal section parsing
                else:
                    if(inLibSection):
                        logging.debug("InLibSection: %s" % line)
                        # get the whole statement library class statement
                        templine = line.strip()
                        while('}' not in templine):
                            i += 1
                            templine += self._RawContent[i].strip()

                        # have good complete line with no whitespace/newline chars
                        # first is the library instance INF
                        # second is the library class

                        libc = templine.partition('{')[2].partition('}')[0].partition(':')[0].strip()
                        libi = templine.partition('{')[0].strip()

                        # Take absolute path and convert to EDK build path
                        RelativePath = pathConverter.GetEdk2RelativePathFromAbsolutePath(libi)
                        if(RelativePath is not None):
                            self.Libraries[libc] = RelativePath
                        else:
                            self.Libraries[libc] = libi
                        i += 1
                        continue

                    elif(inPcdSection):
                        # this is the namespace token line
                        if(len(line.split()) == 1):
                            tokenspace = line

                        # this is the main line of the PCD value
                        elif(line.count("=") == 1 and line.count(":") == 1):
                            while(line.count("\"") % 2) != 0:
                                i += 1
                                line += " " + self._RawContent[i].rstrip()
                            while(line.count('{') != line.count('}')):
                                i += 1
                                line += " " + self._RawContent[i].rstrip()

                            token = line.partition('=')[2]
                            token2 = line.partition(':')[0].split()[-1]
                            self.PCDs[tokenspace + "." + token2] = token.strip()

                        # this is the secondary lines of PCD values showing Defaults
                        elif line.count(":") == 0 and line.count("=") == 1:
                            while(line.count("\"") % 2) != 0:
                                i += 1
                                line += self._RawContent[i].rstrip()

                    elif(inDepSection):
                        pass
                        # not implemented right now

                    else:
                        # not in section...Must be header section
                        line_partitioned = line.partition(':')
                        if(line_partitioned[2] == ""):
                            pass  # not a name: value pair
                        else:
                            key = line_partitioned[0].strip().lower()
                            value = line_partitioned[2].strip()
                            if(key == "module name"):
                                logging.debug("Parsing Mod: %s" % value)
                                self.Name = value
                            elif(key == "module inf path"):
                                self.InfPath = value
                            elif(key == "file guid"):
                                self.Guid = value
                            elif(key == "driver type"):
                                value = value.strip()
                                self.Type = value[value.index('(') + 1:-1]

                i += 1
        except Exception:
            logging.debug("Exception in Parsing: %d" % i)
            raise

#
# Class to parse and objectify the Build report so that
# tools can interact with the Build Report.
# This should simplify the Build Report based interactions
# but should not contain tool specific logic or tests.
#


class BuildReport(object):
    RegionTypes = uf.Enum(['PCD', 'FD', 'MODULE', 'UNKNOWN'])

    def __init__(self, filepath, ws, packagepathcsv, protectedWordsDict):
        self.PlatformName = ""
        self.DscPath = ""
        self.FdfPath = ""
        self.BuildOutputDir = ""
        self.ReportFile = filepath
        self.Modules = {}  # fill this in with objects of ModuleSummary type
        self._ReportContents = ""
        self._Regions = []  # fill this in with tuple (type, start, end)
        self.Workspace = ws  # needs to contain the trailing slash
        self.PackagePathList = []
        for a in packagepathcsv.split(","):
            a = a.strip()
            if(len(a) > 0):
                self.PackagePathList.append(a)
        self.ProtectedWords = protectedWordsDict
        self.PathConverter = pu.Edk2Path(self.Workspace, self.PackagePathList)

    #
    # do region level parsing
    # to get the layout, lists, and dictionaries setup.
    #
    def BasicParse(self):
        if(not os.path.isfile(self.ReportFile)):
            raise Exception("Report File path invalid!")

        # read report
        f = open(self.ReportFile, "r")
        self._ReportContents = [x.strip() for x in f.readlines()]
        f.close()
        #
        # replace protected words
        #
        for (k, v) in self.ProtectedWords.items():
            self._ReportContents = [x.replace(k, v) for x in self._ReportContents]

        logging.debug("Report File is: %s" % self.ReportFile)
        logging.debug("Input report had %d lines of content" % len(self._ReportContents))

        #
        # parse thru and find the regions and basic info at top
        # this is a little hacky in that internal operations could
        # fail but it doesn't seem critical
        #
        linenum = self._GetNextRegionStart(0)
        while(linenum is not None):
            start = linenum
            end = self._GetEndOfRegion(start)
            type = self._GetRegionType(start)
            self._Regions.append((type, start, end))
            linenum = self._GetNextRegionStart(linenum)
            logging.debug("Found a region of type: %s start: %d end: %d" % (BuildReport.RegionTypes[type], start, end))

        #
        # Parse the basic header of the report.
        # we do it after parsing region because we
        # can limit scope to 0 - first start
        #
        for n in range(0, self._Regions[0][1]):  # loop thru from 0 to start of first region
            line = self._ReportContents[n].strip()
            line_partitioned = line.partition(':')
            if(line_partitioned[2] == ""):
                continue

            key = line_partitioned[0].strip().lower()
            value = line_partitioned[2].strip()

            if(key == "platform name"):
                self.PlatformName = value
            elif(key == "platform dsc path"):
                self.DscPath = value
            elif(key == "output path"):
                self.BuildOutputDir = value

        #
        # now for each module summary
        # parse it
        for r in self._Regions:
            if(r[0] == BuildReport.RegionTypes.MODULE):
                mod = ModuleSummary(self._ReportContents[r[1]:r[2]], self.Workspace, self.PackagePathList)
                mod.Parse()
                self.Modules[mod.Guid] = mod

        # now that all modules are parsed lets parse the FD region so we can get the FV name for each module
        for r in self._Regions:
            # if FD region parse out all INFs in the all of the flash
            if(r[0] == BuildReport.RegionTypes.FD):
                self._ParseFdRegionForModules(self._ReportContents[r[1]:r[2]])

    def FindComponentByInfPath(self, InfPath):
        for (k, v) in self.Modules.items():
            if(v.InfPath.lower() == InfPath.lower()):
                logging.debug("Found Module by InfPath: %s" % InfPath)
                return v

        logging.error("Failed to find Module by InfPath %s" % InfPath)
        return None

    def _ParseFdRegionForModules(self, rawcontents):
        FvName = None
        index = 0
        WorkspaceAndPPList = [self.Workspace]
        WorkspaceAndPPList.extend(self.PackagePathList)

        while index < len(rawcontents):
            a = rawcontents[index]
            tokens = a.split()
            if a.startswith("0x") and (len(tokens) == 3) and (a.count('(') == 1):
                if ".inf" not in a.lower() or (a.count('(') != a.count(")")):
                    a = a + rawcontents[index + 1].strip()
                    index += 1
                    tokens = a.split()

                i = a.split()[2].strip().strip('()')

                logging.debug("Found INF in FV Region: " + i)

                # Take absolute path and convert to EDK build path
                RelativePath = self.PathConverter.GetEdk2RelativePathFromAbsolutePath(i)
                if(RelativePath is not None):
                    comp = self.FindComponentByInfPath(RelativePath)
                    if comp is not None:
                        comp.FvName = FvName
                    else:
                        logging.error("Failed to find component for INF path %a" % RelativePath)

            elif a.startswith("Fv Name:"):
                # Fv Name:            FVDXE (99.5% Full)
                FvName = a.partition(":")[2].strip().split()[0]
                logging.debug("Found FvName. RAW: %s  Name: %s" % (a, FvName))
            else:
                logging.debug("ignored line in FD parsing: %s" % a)
            index += 1

        return

    #
    # Get the start of region
    #
    def _GetNextRegionStart(self, number):
        lineNumber = number
        while(lineNumber < len(self._ReportContents)):
            if self._ReportContents[lineNumber] == ">======================================================================================================================<":  # noqa: E501
                return lineNumber + 1
            lineNumber += 1
        logging.debug("Failed to find a Start Next Region after lineNumber: %d" % number)
        # didn't find new region
        return None

    #
    # Get the end of region
    #
    def _GetEndOfRegion(self, number):
        lineNumber = number
        while(lineNumber < len(self._ReportContents)):
            if self._ReportContents[lineNumber] == "<======================================================================================================================>":  # noqa: E501
                return lineNumber - 1
            lineNumber += 1

        logging.debug("Failed to find a End Region after lineNumber: %d" % number)
        # didn't find new region
        return None

    def _GetRegionType(self, lineNumber):
        line = self._ReportContents[lineNumber].strip()
        if(line == "Firmware Device (FD)"):
            return BuildReport.RegionTypes.FD
        elif(line == "Platform Configuration Database Report"):
            return BuildReport.RegionTypes.PCD
        elif(line == "Module Summary"):
            return BuildReport.RegionTypes.MODULE
        else:
            return BuildReport.RegionTypes.UNKNOWN



# @file tpm2_simulator.py
# This file contains transportation layer classes for interacting with the TPM 2.0 simulator.
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import socket
import struct
import edk2toollib.tpm.tpm2_stream as t2s
import edk2toollib.tpm.tpm2_defs as t2d

PLAT_COMMANDS = {
    'TPM_SIGNAL_POWER_ON': 1,
    'TPM_SIGNAL_POWER_OFF': 2,
    'TPM_SIGNAL_PHYS_PRES_ON': 3,
    'TPM_SIGNAL_PHYS_PRES_OFF': 4,
    'TPM_SIGNAL_HASH_START': 5,
    'TPM_SIGNAL_HASH_DATA': 6,
    # {UINT32 BufferSize, BYTE[BufferSize] Buffer}
    'TPM_SIGNAL_HASH_END': 7,
    'TPM_SEND_COMMAND': 8,
    # {BYTE Locality, UINT32 InBufferSize, BYTE[InBufferSize] InBuffer} ->
    # {UINT32 OutBufferSize, BYTE[OutBufferSize] OutBuffer}
    'TPM_SIGNAL_CANCEL_ON': 9,
    'TPM_SIGNAL_CANCEL_OFF': 10,
    'TPM_SIGNAL_NV_ON': 11,
    'TPM_SIGNAL_NV_OFF': 12,
    'TPM_SIGNAL_KEY_CACHE_ON': 13,
    'TPM_SIGNAL_KEY_CACHE_OFF': 14,
    'TPM_REMOTE_HANDSHAKE': 15,
    'TPM_SET_ALTERNATIVE_RESULT': 16,
    'TPM_SIGNAL_RESET': 17,
    'TPM_SESSION_END': 20,
    'TPM_STOP': 21,
    'TPM_GET_COMMAND_RESPONSE_SIZES': 25,
    'TPM_TEST_FAILURE_MODE': 30,
}


class TpmSimulator(object):

    def __init__(self, host='localhost', port=2321):
        super(TpmSimulator, self).__init__()

        # Connect to the control socket.
        self.platSock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.platSock.connect((host, port + 1))

        # Connect to the simulator socket.
        self.tpmSock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.tpmSock.connect((host, port))

        # Power cycle the TPM.
        self.platSock.send(struct.pack(">L", PLAT_COMMANDS['TPM_SIGNAL_POWER_OFF']))
        self.platSock.send(struct.pack(">L", PLAT_COMMANDS['TPM_SIGNAL_POWER_ON']))

        # Enable the NV space.
        self.platSock.send(struct.pack(">L", PLAT_COMMANDS['TPM_SIGNAL_NV_ON']))

    def send_raw_data(self, data):
        print("RAW -->: " + str(data).encode('hex'))
        self.tpmSock.send(data)

    def read_raw_data(self, count):
        data = self.tpmSock.recv(count)
        print("RAW <--: " + str(data).encode('hex'))
        return data

    def send_data(self, data):
        # Send the "I'm about to send data" command.
        self.send_raw_data(struct.pack(">L", PLAT_COMMANDS['TPM_SEND_COMMAND']))
        # Send the locality for the data.
        self.send_raw_data(struct.pack(">b", 0x03))
        # Send the size of the data.
        self.send_raw_data(struct.pack(">L", len(data)))

        # Now, send the data itself.
        self.send_raw_data(data)

        # Poll until a result is available.
        # NOTE: This shouldn't be necessary and denotes a lack of understanding...
        while True:
            result_size = self.read_raw_data(4)
            result_size = struct.unpack(">L", result_size)[0]
            if (result_size > 0):
                break

        return self.read_raw_data(result_size)

    def startup(self, type):
        stream = t2s.Tpm2CommandStream(t2d.TPM_ST_NO_SESSIONS, 0x00, t2d.TPM_CC_Startup)
        stream.add_element(t2s.Tpm2StreamPrimitive(t2d.TPM_SU_Size, type))
        return self.send_data(stream.get_stream())



# @file tpm2_stream.py
# This file contains utility classes to help marshal and unmarshal data to/from the TPM.
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##


import struct


class Tpm2StreamElement(object):
    def __init__(self):
        self.pack_string = ""

    """This get_size refers to the size of this structure when marshalled"""

    def get_size(self):
        return struct.calcsize(self.pack_string)


class Tpm2StreamPrimitive(Tpm2StreamElement):
    def __init__(self, size, value):
        super(Tpm2StreamPrimitive, self).__init__()

        if size not in (1, 2, 4, 8):
            raise ValueError("Size must be 1, 2, 4, or 8 bytes!")

        self.pack_string = {
            1: ">B",
            2: ">H",
            4: ">L",
            8: ">Q"
        }[size]
        self.value = value

    def marshal(self):
        return struct.pack(self.pack_string, self.value)


class TPM2_COMMAND_HEADER(Tpm2StreamElement):
    def __init__(self, tag, size, code):
        super(TPM2_COMMAND_HEADER, self).__init__()
        self.tag = tag
        self.code = code
        self.size = size
        self.pack_string = ">HLL"

    """This update_size refers to the size of the whole command"""

    def update_size(self, size):
        self.size = size

    def marshal(self):
        return struct.pack(self.pack_string, self.tag, self.size, self.code)


class TPM2B(Tpm2StreamElement):
    def __init__(self, data):
        super(TPM2B, self).__init__()
        self.data = data
        self.size = len(data)
        self.pack_string = ">H%ds" % self.size

    def update_data(self, data):
        self.data = data
        self.size = len(data)
        self.pack_string = ">H%ds" % self.size

    def marshal(self):
        return struct.pack(self.pack_string, self.size, self.data)


class Tpm2CommandStream(object):
    def __init__(self, tag, size, code):
        super(Tpm2CommandStream, self).__init__()
        self.header = TPM2_COMMAND_HEADER(tag, size, code)
        self.stream_size = self.header.get_size()
        self.header.update_size(self.stream_size)
        self.stream_elements = []

    def get_size(self):
        return self.stream_size

    def add_element(self, element):
        self.stream_elements.append(element)
        self.stream_size += element.get_size()
        self.header.update_size(self.stream_size)

    def get_stream(self):
        return self.header.marshal() + b''.join(element.marshal() for element in self.stream_elements)



# @file
# Code to help convert an Int to StatusCode string
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##


class UefiStatusCode(object):
    # string Array
    StatusCodeStrings = ["Success", "Load Error", "Invalid Parameter", "Unsupported", "Bad BufferSize",
                         "Buffer Too Small", "Not Ready", "Device Error", "Write Protected", "Out of Resources",
                         "Volume Corrupt", "Volume Full", "No Media", "Media Changed", "Not Found", "Access Denied",
                         "No Response", "No Mapping", "Time Out", "Not Started", "Already Started", "Aborted",
                         "ICMP Error", "TFTP Error", "Protocol Error", "Incompatible Error", "Security Violation",
                         "CRC Error", "End of Media", "Reserved(29)", "Reserved(30)", "End of File",
                         "Invalid Language", "Compromised Data"]

    def Convert32BitToString(self, i):
        # convert a 32bit value to string
        return UefiStatusCode.StatusCodeStrings[(i & 0xFFF)]

    def Convert64BitToString(self, l):
        if(l > len(UefiStatusCode.StatusCodeStrings)):
            return ""
        return UefiStatusCode.StatusCodeStrings[(l & 0xFFF)]

    def ConvertHexString64ToString(self, hexstring):
        value = int(hexstring, 16)
        return self.Convert64BitToString(value)

    def ConvertHexString32ToString(self, hexstring):
        value = int(hexstring, 16)
        return self.Convert32BitToString(value)



## @file
# Script to generate Cat files for capsule update based on supplied inf file.
# This uses the winsdk and the command line tool Inf2Cat.exe
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import os
import logging
from edk2toollib.utility_functions import RunCmd
from edk2toollib.windows.locate_tools import FindToolInWinSdk


class CatGenerator(object):
    SUPPORTED_OS = {'win10': '10',
                    '10': '10',
                    '10_au': '10_AU',
                    '10_rs2': '10_RS2',
                    '10_rs3': '10_RS3',
                    '10_rs4': '10_RS4',
                    'server10': 'Server10',
                    'server2016': 'Server2016',
                    'serverrs2': 'ServerRS2',
                    'serverrs3': 'ServerRS3',
                    'serverrs4': 'ServerRS4'
                    }

    def __init__(self, arch, os):
        self.Arch = arch
        self.OperatingSystem = os

    @property
    def Arch(self):
        return self._arch

    @Arch.setter
    def Arch(self, value):
        value = value.lower()
        if(value == "x64") or (value == "amd64"):  # support amd64 value so INF and CAT tools can use same arch value
            self._arch = "X64"
        elif(value == "arm"):
            self._arch = "ARM"
        elif(value == "arm64") or (value == "aarch64"):  # support UEFI defined aarch64 value as well
            self._arch = "ARM64"
        else:
            logging.critical("Unsupported Architecture: %s", value)
            raise ValueError("Unsupported Architecture")

    @property
    def OperatingSystem(self):
        return self._operatingsystem

    @OperatingSystem.setter
    def OperatingSystem(self, value):
        key = value.lower()
        if(key not in CatGenerator.SUPPORTED_OS.keys()):
            logging.critical("Unsupported Operating System: %s", key)
            raise ValueError("Unsupported Operating System")
        self._operatingsystem = CatGenerator.SUPPORTED_OS[key]

    def MakeCat(self, OutputCatFile, PathToInf2CatTool=None):
        # Find Inf2Cat tool
        if(PathToInf2CatTool is None):
            PathToInf2CatTool = FindToolInWinSdk("Inf2Cat.exe")
        # check if exists
        if PathToInf2CatTool is None or not os.path.exists(PathToInf2CatTool):
            raise Exception("Can't find Inf2Cat on this machine.  Please install the Windows 10 WDK - "
                            "https://developer.microsoft.com/en-us/windows/hardware/windows-driver-kit")

        OutputFolder = os.path.dirname(OutputCatFile)
        # Make Cat file
        cmd = "/driver:. /os:" + self.OperatingSystem + "_" + self.Arch + " /verbose"
        ret = RunCmd(PathToInf2CatTool, cmd, workingdir=OutputFolder)
        if(ret != 0):
            raise Exception("Creating Cat file Failed with errorcode %d" % ret)
        if(not os.path.isfile(OutputCatFile)):
            raise Exception("CAT file (%s) not created" % OutputCatFile)

        return 0



# @file dsc_parser.py
# Code to help parse DSC files
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
from edk2toollib.uefi.edk2.parsers.base_parser import HashFileParser
import os


class DscParser(HashFileParser):

    def __init__(self):
        super(DscParser, self).__init__('DscParser')
        self.SixMods = []
        self.SixModsEnhanced = []
        self.ThreeMods = []
        self.ThreeModsEnhanced = []
        self.OtherMods = []
        self.Libs = []
        self.LibsEnhanced = []
        self.ParsingInBuildOption = 0
        self.LibraryClassToInstanceDict = {}
        self.Pcds = []

    def __ParseLine(self, Line, file_name=None, lineno=None):
        line_stripped = self.StripComment(Line).strip()
        if(len(line_stripped) < 1):
            return ("", [], None)

        line_resolved = self.ReplaceVariables(line_stripped)
        if(self.ProcessConditional(line_resolved)):
            # was a conditional
            # Other parser returns line_resolved, [].  Need to figure out which is right
            return ("", [], None)

        # not conditional keep procesing

        # check if conditional is active
        if(not self.InActiveCode()):
            return ("", [], None)

        # check for include file and import lines from file
        if(line_resolved.strip().lower().startswith("!include")):
            # include line.
            toks = line_resolved.split()
            self.Logger.debug("Opening Include File %s" % os.path.join(self.RootPath, toks[1]))
            sp = self.FindPath(toks[1])
            lf = open(sp, "r")
            loc = lf.readlines()
            lf.close()
            return ("", loc, sp)

        # check for new section
        (IsNew, Section) = self.ParseNewSection(line_resolved)
        if(IsNew):
            self.CurrentSection = Section.upper()
            self.Logger.debug("New Section: %s" % self.CurrentSection)
            self.Logger.debug("FullSection: %s" % self.CurrentFullSection)
            return (line_resolved, [], None)

        # process line in x64 components
        if(self.CurrentFullSection.upper() == "COMPONENTS.X64"):
            if(self.ParsingInBuildOption > 0):
                if(".inf" in line_resolved.lower()):
                    p = self.ParseInfPathLib(line_resolved)
                    self.Libs.append(p)
                    self.Logger.debug("Found Library in a 64bit BuildOptions Section: %s" % p)
                elif "tokenspaceguid" in line_resolved.lower() and \
                        line_resolved.count('|') > 0 and line_resolved.count('.') > 0:
                    # should be a pcd statement
                    p = line_resolved.partition('|')
                    self.Pcds.append(p[0].strip())
                    self.Logger.debug("Found a Pcd in a 64bit Module Override section: %s" % p[0].strip())
            else:
                if(".inf" in line_resolved.lower()):
                    p = self.ParseInfPathMod(line_resolved)
                    self.SixMods.append(p)
                    if file_name is not None and lineno is not None:
                        self.SixModsEnhanced.append({'file': os.path.normpath(file_name), 'lineno': lineno, 'data': p})
                    self.Logger.debug("Found 64bit Module: %s" % p)

            self.ParsingInBuildOption = self.ParsingInBuildOption + line_resolved.count("{")
            self.ParsingInBuildOption = self.ParsingInBuildOption - line_resolved.count("}")
            return (line_resolved, [], None)

        # process line in ia32 components
        elif(self.CurrentFullSection.upper() == "COMPONENTS.IA32"):
            if(self.ParsingInBuildOption > 0):
                if(".inf" in line_resolved.lower()):
                    p = self.ParseInfPathLib(line_resolved)
                    self.Libs.append(p)
                    if file_name is not None and lineno is not None:
                        self.LibsEnhanced.append({'file': os.path.normpath(file_name), 'lineno': lineno, 'data': p})
                    self.Logger.debug("Found Library in a 32bit BuildOptions Section: %s" % p)
                elif "tokenspaceguid" in line_resolved.lower() and \
                        line_resolved.count('|') > 0 and line_resolved.count('.') > 0:
                    # should be a pcd statement
                    p = line_resolved.partition('|')
                    self.Pcds.append(p[0].strip())
                    self.Logger.debug("Found a Pcd in a 32bit Module Override section: %s" % p[0].strip())

            else:
                if(".inf" in line_resolved.lower()):
                    p = self.ParseInfPathMod(line_resolved)
                    self.ThreeMods.append(p)
                    if file_name is not None and lineno is not None:
                        self.ThreeModsEnhanced.append({'file': os.path.normpath(file_name),
                                                       'lineno': lineno, 'data': p})
                    self.Logger.debug("Found 32bit Module: %s" % p)

            self.ParsingInBuildOption = self.ParsingInBuildOption + line_resolved.count("{")
            self.ParsingInBuildOption = self.ParsingInBuildOption - line_resolved.count("}")
            return (line_resolved, [], None)

        # process line in other components
        elif("COMPONENTS" in self.CurrentFullSection.upper()):
            if(self.ParsingInBuildOption > 0):
                if(".inf" in line_resolved.lower()):
                    p = self.ParseInfPathLib(line_resolved)
                    self.Libs.append(p)
                    self.Logger.debug("Found Library in a BuildOptions Section: %s" % p)
                elif "tokenspaceguid" in line_resolved.lower() and \
                        line_resolved.count('|') > 0 and line_resolved.count('.') > 0:
                    # should be a pcd statement
                    p = line_resolved.partition('|')
                    self.Pcds.append(p[0].strip())
                    self.Logger.debug("Found a Pcd in a Module Override section: %s" % p[0].strip())

            else:
                if(".inf" in line_resolved.lower()):
                    p = self.ParseInfPathMod(line_resolved)
                    self.OtherMods.append(p)
                    self.Logger.debug("Found Module: %s" % p)

            self.ParsingInBuildOption = self.ParsingInBuildOption + line_resolved.count("{")
            self.ParsingInBuildOption = self.ParsingInBuildOption - line_resolved.count("}")
            return (line_resolved, [], None)

        # process line in library class section (don't use full name)
        elif(self.CurrentSection.upper() == "LIBRARYCLASSES"):
            if(".inf" in line_resolved.lower()):
                p = self.ParseInfPathLib(line_resolved)
                self.Libs.append(p)
                self.Logger.debug("Found Library in Library Class Section: %s" % p)
            return (line_resolved, [], None)
        # process line in PCD section
        elif(self.CurrentSection.upper().startswith("PCDS")):
            if "tokenspaceguid" in line_resolved.lower() and \
                    line_resolved.count('|') > 0 and line_resolved.count('.') > 0:
                # should be a pcd statement
                p = line_resolved.partition('|')
                self.Pcds.append(p[0].strip())
                self.Logger.debug("Found a Pcd in a PCD section: %s" % p[0].strip())
            return (line_resolved, [], None)
        else:
            return (line_resolved, [], None)

    def __ParseDefineLine(self, Line):
        line_stripped = self.StripComment(Line).strip()
        if(len(line_stripped) < 1):
            return ("", [])

        # this line needs to be here to resolve any symbols inside the !include lines, if any
        line_resolved = self.ReplaceVariables(line_stripped)
        if(self.ProcessConditional(line_resolved)):
            # was a conditional
            # Other parser returns line_resolved, [].  Need to figure out which is right
            return ("", [])

        # not conditional keep procesing

        # check if conditional is active
        if(not self.InActiveCode()):
            return ("", [])

        # check for include file and import lines from file
        if(line_resolved.strip().lower().startswith("!include")):
            # include line.
            toks = line_resolved.split()
            self.Logger.debug("Opening Include File %s" % os.path.join(self.RootPath, toks[1]))
            sp = self.FindPath(toks[1])
            lf = open(sp, "r")
            loc = lf.readlines()
            lf.close()
            return ("", loc)

        # check for new section
        (IsNew, Section) = self.ParseNewSection(line_resolved)
        if(IsNew):
            self.CurrentSection = Section.upper()
            self.Logger.debug("New Section: %s" % self.CurrentSection)
            self.Logger.debug("FullSection: %s" % self.CurrentFullSection)
            return (line_resolved, [])

        # process line based on section we are in
        if(self.CurrentSection == "DEFINES") or (self.CurrentSection == "BUILDOPTIONS"):
            if line_resolved.count("=") >= 1:
                tokens = line_resolved.split("=", 1)
                leftside = tokens[0].split()
                if(len(leftside) == 2):
                    left = leftside[1]
                else:
                    left = leftside[0]
                right = tokens[1].strip()

                self.LocalVars[left] = right
                self.Logger.debug("Key,values found:  %s = %s" % (left, right))

                # iterate through the existed LocalVars and try to resolve the symbols
                for var in self.LocalVars:
                    self.LocalVars[var] = self.ReplaceVariables(self.LocalVars[var])
                return (line_resolved, [])
        else:
            return (line_resolved, [])

    def ParseInfPathLib(self, line):
        if(line.count("|") > 0):
            line_parts = []
            c = line.split("|")[0].strip()
            i = line.split("|")[1].strip()
            if(c in self.LibraryClassToInstanceDict):
                line_parts = self.LibraryClassToInstanceDict.get(c)
            sp = self.FindPath(i)
            line_parts.append(sp)
            self.LibraryClassToInstanceDict[c] = line_parts
            return line.split("|")[1].strip()
        else:
            return line.strip().split()[0]

    def ParseInfPathMod(self, line):
        return line.strip().split()[0].rstrip("{")

    def __ProcessMore(self, lines, file_name=None):
        if(len(lines) > 0):
            for index in range(0, len(lines)):
                (line, add, new_file) = self.__ParseLine(lines[index], file_name=file_name, lineno=index + 1)
                if(len(line) > 0):
                    self.Lines.append(line)
                self.__ProcessMore(add, file_name=new_file)

    def __ProcessDefines(self, lines):
        if(len(lines) > 0):
            for l in lines:
                (line, add) = self.__ParseDefineLine(l)
                self.__ProcessDefines(add)

    def ResetParserState(self):
        #
        # add more DSC parser based state reset here, if necessary
        #
        super(DscParser, self).ResetParserState()

    def ParseFile(self, filepath):
        self.Logger.debug("Parsing file: %s" % filepath)
        self.TargetFile = os.path.abspath(filepath)
        self.TargetFilePath = os.path.dirname(self.TargetFile)
        f = open(os.path.join(filepath), "r")
        # expand all the lines and include other files
        file_lines = f.readlines()
        self.__ProcessDefines(file_lines)
        # reset the parser state before processing more
        self.ResetParserState()
        self.__ProcessMore(file_lines, file_name=os.path.join(filepath))
        f.close()
        self.Parsed = True

    def GetMods(self):
        return self.ThreeMods + self.SixMods

    def GetModsEnhanced(self):
        return self.ThreeModsEnhanced + self.SixModsEnhanced

    def GetLibs(self):
        return self.Libs

    def GetLibsEnhanced(self):
        return self.LibsEnhanced



# @file tpm2_defs_test.py
# This file contains utility classes to help interpret definitions from the
# Tpm20.h header file in TianoCore.
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import unittest
import edk2toollib.tpm.tpm2_defs as t2d


class TestCommandCode(unittest.TestCase):

    def test_get_code_returns_codes(self):
        self.assertEqual(t2d.CommandCode.get_code('TPM_CC_Clear'), 0x00000126)
        self.assertEqual(t2d.CommandCode.get_code('TPM_CC_ActivateCredential'), 0x00000147)

    def test_get_code_returns_none_if_not_found(self):
        self.assertEqual(t2d.CommandCode.get_code('I_AM_NOT_A_VALID_CODE'), None)
        self.assertEqual(t2d.CommandCode.get_code(None), None)

    def test_get_string_returns_strings(self):
        self.assertEqual(t2d.CommandCode.get_string(0x00000126), 'TPM_CC_Clear')
        self.assertEqual(t2d.CommandCode.get_string(0x00000147), 'TPM_CC_ActivateCredential')

    def test_get_string_returns_none_if_not_found(self):
        self.assertEqual(t2d.CommandCode.get_string(0xFFFFFFFF), None)



# @file locate_tools.py
# This module provides python services that locate common development tools using vswhere.exe,
# vsvars.bat, and other Windows based tools.  This works best on systems running Windows
# and with dev tools installed but will attempt to use known paths for WinSDK if the dev
# tools are not available.  This is only a best effort to locate the SDK tools in well
# known/default install locations.
#
# Suggested Dev Tools:
#   Current Windows SDKs
#   Visual Studio 2017 Build Tools or newer
#
# Note: When this module is used in local develop mode it will download vswhere.exe from the github.
#       It will confirm the hash is a known good value.
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import pkg_resources
import os
import logging
import glob
import subprocess
from edk2toollib.utility_functions import RunCmd
from edk2toollib.utility_functions import GetHostInfo
import re
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
import urllib.error
import urllib.request

# Update this when you want a new version of VsWhere
__VERSION = "2.6.7"
__URL = "https://github.com/microsoft/vswhere/releases/download/{}/vswhere.exe".format(__VERSION)
__SHA256 = "10abd21aeb5003d87c01f033fd7c170360e362be397f23b0b730324abbd92612"

#
# Supported Versions that can be queried with vswhere
# Use lower case for key as all comparisons will be lower case
#
supported_vs_versions = {"vs2017": "15.0,16.0", "vs2019": "16.0,17.0"}


# Downloads VSWhere
def _DownloadVsWhere(unpack_folder: os.PathLike = None):
    if unpack_folder is None:
        unpack_folder = os.path.dirname(__VsWherePath())

    out_file_name = os.path.join(unpack_folder, "vswhere.exe")
    logging.info("Attempting to download vswhere to: {}. This may take a second.".format(unpack_folder))
    # check if we have the vswhere file already downloaded
    if not os.path.isfile(out_file_name):
        try:
            # Download the file and save it locally under `temp_file_name`
            with urllib.request.urlopen(__URL) as response, open(out_file_name, 'wb') as out_file:
                out_file.write(response.read())
        except urllib.error.HTTPError as e:
            logging.error(f"We ran into an issue when getting VsWhere")
            raise e

    # do the hash to make sure the file is good
    with open(out_file_name, "rb") as file:
        import hashlib
        temp_file_sha256 = hashlib.sha256(file.read()).hexdigest()
    if temp_file_sha256 != __SHA256:
        # delete the file since it's not what we're expecting
        os.remove(out_file_name)
        raise RuntimeError(f"VsWhere - sha256 does not match\n\tdownloaded:\t{temp_file_sha256}\n\t")


def __VsWherePath():
    file = "vswhere.exe"
    requirement = pkg_resources.Requirement.parse("edk2-pytool-library")
    file_path = os.path.join("edk2toollib", "bin", file)
    vswhere_path = pkg_resources.resource_filename(requirement, file_path)
    return vswhere_path


####
#
# https://docs.microsoft.com/en-us/vswhere/install-vswhere-client-tools
#
# @return string "/PATH/TO/vswhere.exe" or None
####
def GetVsWherePath(fail_on_not_found: bool = True):
    vswhere_path = __VsWherePath()
    # check if we can't find it, look for vswhere in the path
    if not os.path.isfile(vswhere_path):
        for env_var in os.getenv("PATH").split(os.pathsep):
            env_var = os.path.join(os.path.normpath(env_var), "vswhere.exe")
            if os.path.isfile(env_var):
                vswhere_path = env_var
                break

    # if we still can't find it, download it
    if not os.path.isfile(vswhere_path):
        vswhere_dir = os.path.dirname(vswhere_path)
        try:  # try to download
            _DownloadVsWhere(vswhere_dir)
        except Exception:
            logging.warning("Tried to download VsWhere and failed")
            pass

    # if we're still hosed
    if not os.path.isfile(vswhere_path) and fail_on_not_found:
        logging.error("We weren't able to find vswhere!")
        return None

    return vswhere_path


####
# Finds a product with VS Where
#
# product: is defined by vswhere tool
# vs_version: helper to find version of supported VS version (example vs2019).
####
def FindWithVsWhere(products: str = "*", vs_version: str = None):
    cmd = "-latest -nologo -all -property installationPath"
    vs_where_path = GetVsWherePath()
    if vs_where_path is None:
        logging.warning("We weren't able to find VSWhere")
        return (1, None)
    if(products is not None):
        cmd += " -products " + products
    if(vs_version is not None):
        vs_version = vs_version.lower()
        if vs_version in supported_vs_versions.keys():
            cmd += " -version " + supported_vs_versions[vs_version]
        else:
            logging.warning("Invalid or unsupported vs_version " + vs_version)
            return (2, None)
    a = StringIO()
    ret = RunCmd(vs_where_path, cmd, outstream=a)
    if(ret != 0):
        a.close()
        return (ret, None)
    p1 = a.getvalue().strip()
    a.close()
    if(len(p1.strip()) > 0):
        return (0, p1)
    return (ret, None)


# Run visual studio batch file and collect the
# interesting environment values
#
#  Inspiration taken from cpython for this method of env collection
#
# keys: enumerable list with names of env variables to collect after bat run
# arch: arch to run.  amd64, x86, ??
# product: value defined by vswhere.exe
# vs_version: helper to find version of supported VS version (example vs2019).
# returns a dictionary of the interesting environment variables
def QueryVcVariables(keys: list, arch: str = None, product: str = None, vs_version: str = None):
    """Launch vcvarsall.bat and read the settings from its environment.  This is a windows only function
    and Windows is case insensitive for the keys"""

    if product is None:
        product = "*"
    if arch is None:
        # TODO: look up host architecture?
        arch = "amd64"
    interesting = set(x.upper() for x in keys)
    result = {}
    ret, vs_path = FindWithVsWhere(product, vs_version)
    if ret != 0 or vs_path is None:
        logging.warning("We didn't find VS path or otherwise failed to invoke vsWhere")
        raise ValueError("Bad VC")
    vcvarsall_path = os.path.join(vs_path, "VC", "Auxiliary", "Build", "vcvarsall.bat")
    logging.debug("Calling '%s %s'", vcvarsall_path, arch)
    popen = subprocess.Popen('"%s" %s & set' % (vcvarsall_path, arch), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    try:
        stdout, stderr = popen.communicate()
        if popen.wait() != 0:
            raise Exception(stderr.decode("mbcs"))
        stdout = stdout.decode("mbcs")
        for line in stdout.split("\n"):
            if '=' not in line:
                continue
            line = line.strip()
            key, value = line.split('=', 1)
            if key.upper() in interesting:
                if value.endswith(os.pathsep):
                    value = value[:-1]
                result[key] = value
    finally:
        popen.stdout.close()
        popen.stderr.close()

    if len(result) != len(interesting):
        logging.debug("Input: " + str(sorted(interesting)))
        logging.debug("Result: " + str(sorted(list(result.keys()))))
        raise ValueError(str(list(result.keys())))
    return result


# return 1 if a > b
# return 0 if b == a
# return -1 if a < b
def _CompareWindowVersions(a, b):
    a_periods = str(a).count(".")
    b_periods = str(b).count(".")
    if a_periods == 3 and b_periods != 3:
        return 1
    if b_periods == 3 and a_periods != 3:
        return -1
    if a_periods != 3 and b_periods != 3:
        return 0
    a_parts = str(a).split(".")
    b_parts = str(b).split(".")
    for i in range(3):
        a_p = int(a_parts[i])
        b_p = int(b_parts[i])
        if a_p > b_p:
            return 1
        if a_p < b_p:
            return -1
    return 0


def _CheckArchOfMatch(match):
    '''
    Returns if this binary matches our host
    returns true or false
    if no arch is in the match, then we return true
    '''
    match = str(match).lower()
    isx86 = "x86" in match
    isx64 = "x64" in match or "amd64" in match
    isArm64 = "aarch" in match or "aarch64" in match or "arm64" in match
    isi386 = "i386" in match
    isArm = not isArm64 and ("arm" in match)
    count = 0
    count += 1 if isx64 else 0
    count += 1 if isx86 else 0
    count += 1 if isArm else 0
    count += 1 if isArm64 else 0
    count += 1 if isi386 else 0
    if count == 0:  # we don't know what arch this is?
        return True
    if count > 1:  # there are more than one arch for this binary
        logging.warning("We found more than one architecture for {}. Results maybe inconsistent".format(match))
        return True

    _, arch, bits = GetHostInfo()
    bits = int(bits)
    if isx86 and (bits < 32 or arch != "x86"):
        return False
    if isx64 and (bits < 64 or arch != "x86"):
        return False
    if isi386:
        # TODO add i386 to GetHostInfo
        return False
    if isArm64 and (bits < 64 or arch != "ARM"):
        return False
    if isArm and (bits < 32 or arch != "ARM"):
        return False
    return True


# does a glob in the folder that your sdk is
# uses the environmental variable WindowsSdkDir and tries to use WindowsSDKVersion
def FindToolInWinSdk(tool: str, product=None, arch=None):
    variables = ["WindowsSdkDir", "WindowsSDKVersion"]
    # get the value with QueryVcVariables
    try:
        results = QueryVcVariables(variables, product, arch)
        # Get the variables we care about
        sdk_dir = results["WindowsSdkDir"]
        sdk_ver = results["WindowsSDKVersion"]
    except ValueError:
        sdk_dir = os.path.join(os.getenv("ProgramFiles(x86)"), "Windows Kits", "10", "bin")
        sdk_ver = "0.0.0.0"
    sdk_dir = os.path.realpath(sdk_dir)
    search_pattern = os.path.join(sdk_dir, "**", tool)

    match_offset = len(sdk_dir)
    # look for something like 10.0.12323.0123
    windows_ver_regex = re.compile(r'\d+\.\d+\.\d+\.\d+')
    top_version_so_far = -1
    top_match = None
    # Look at in match in the tree
    for match in glob.iglob(search_pattern, recursive=True):
        match_file = match[match_offset:]  # strip off the root
        match_front, match_end = os.path.split(match_file)  # split off the filename
        versions = windows_ver_regex.findall(match_front)  # look for windows versions
        top_ver_match = 0
        if not _CheckArchOfMatch(match_front):  # make sure it's a good arch for us
            continue
        if len(versions) == 0:
            top_ver_match = "0.0.0.0"  # if we have a bad version, we should fall to a bad version
        for version in versions:  # find the highest version if there are multiple in this?
            is_current_sdk_version = _CompareWindowVersions(version, sdk_ver) == 0
            if _CompareWindowVersions(version, top_ver_match) > 0 or is_current_sdk_version:
                top_ver_match = version
        # if we have a highest version or one that matches our current from environment variables?
        is_current_sdk_version = _CompareWindowVersions(top_ver_match, sdk_ver) == 0
        if _CompareWindowVersions(top_ver_match, top_version_so_far) > 0 or is_current_sdk_version:
            top_version_so_far = top_ver_match
            top_match = match
    if top_match is None:
        logging.critical("We weren't able to find {}".format(tool))
    return top_match



# @file
# Module contains helper classes and functions to work with UEFI FFs.
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import uuid
import struct
import sys

#
# EFI_FFS_FILE_HEADER
#
# typedef struct {
#   EFI_GUID                Name;
#   EFI_FFS_INTEGRITY_CHECK IntegrityCheck;
#   EFI_FV_FILETYPE         Type;
#   EFI_FFS_FILE_ATTRIBUTES Attributes;
#   UINT8                   Size[3];
#   EFI_FFS_FILE_STATE      State;
# } EFI_FFS_FILE_HEADER;


class EfiFirmwareFileSystemHeader(object):
    def __init__(self):
        self.StructString = "=16sHBBBBBB"
        self.FileSystemGuid = None
        self.Size0 = None
        self.Size1 = None
        self.Size2 = None
        self.Attributes = None
        self.Type = None
        self.State = None

    def get_size(self):
        return self.Size0 + (self.Size1 << 8) + (self.Size2 << 16)

    def load_from_file(self, file):
        orig_seek = file.tell()
        struct_bytes = file.read(struct.calcsize(self.StructString))
        file.seek(orig_seek)

        # Load this object with the contents of the data.
        (self.FileSystemGuid, self.Checksum, self.Type, self.Attributes, self.Size0, self.Size1,
            self.Size2, self.State) = struct.unpack(self.StructString, struct_bytes)

        # Update the GUID to be a UUID object.
        if sys.byteorder == 'big':
            self.FileSystemGuid = uuid.UUID(bytes=self.FileSystemGuid)
        else:
            self.FileSystemGuid = uuid.UUID(bytes_le=self.FileSystemGuid)

        return self

    def serialize(self):
        file_system_guid_bin = self.FileSystemGuid.bytes if sys.byteorder == 'big' else self.FileSystemGuid.bytes_le
        return struct.pack(self.StructString, file_system_guid_bin, self.Checksum,
                           self.Type, self.Attributes, self.Size0, self.Size1, self.Size2, self.State)



##
# Utility Functions to support re-use in python scripts.
# Includes functions for running external commands, etc
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import re
import os
import logging
import datetime
import time
import shutil
import threading
import subprocess
import sys
import inspect
import platform
import importlib
from collections import namedtuple

####
# Helper to allow Enum type to be used which allows better code readability
#
# ref: http://stackoverflow.com/questions/36932/how-can-i-represent-an-enum-in-python
####


class Enum(tuple):
    __getattr__ = tuple.index


####
# Class to support running commands from the shell in a python environment.
# Don't use directly.
#
# PropagatingThread copied from sample here:
# https://stackoverflow.com/questions/2829329/catch-a-threads-exception-in-the-caller-thread-in-python
####
class PropagatingThread(threading.Thread):
    def run(self):
        self.exc = None
        try:
            if hasattr(self, '_Thread__target'):
                # Thread uses name mangling prior to Python 3.
                self.ret = self._Thread__target(*self._Thread__args, **self._Thread__kwargs)
            else:
                self.ret = self._target(*self._args, **self._kwargs)
        except BaseException as e:
            self.exc = e

    def join(self, timeout=None):
        super(PropagatingThread, self).join()
        if self.exc:
            raise self.exc
        return self.ret


####
# Helper functions for running commands from the shell in python environment
# Don't use directly
#
# process output stream and write to log.
# part of the threading pattern.
#
#  http://stackoverflow.com/questions/19423008/logged-subprocess-communicate
####
def reader(filepath, outstream, stream, logging_level=logging.INFO):
    f = None
    # open file if caller provided path
    if(filepath):
        f = open(filepath, "w")

    while True:
        s = stream.readline().decode()
        if not s:
            break
        if(f is not None):
            # write to file if caller provided file
            f.write(s)
        if(outstream is not None):
            # write to stream object if caller provided object
            outstream.write(s)
        logging.log(logging_level, s.rstrip())
    stream.close()
    if(f is not None):
        f.close()


####
# Returns a namedtuple containing information about host machine.
#
# @return namedtuple Host(os=OS Type, arch=System Architecture, bit=Highest Order Bit)
####
def GetHostInfo():
    Host = namedtuple('Host', 'os arch bit')
    host_info = platform.uname()
    os = host_info.system
    processor_info = host_info.machine
    logging.debug("Getting host info for host: {0}".format(str(host_info)))

    arch = None
    bit = None

    if ("x86" in processor_info) or ("AMD" in processor_info) or ("Intel" in processor_info):
        arch = "x86"
    elif ("ARM" in processor_info) or ("AARCH" in processor_info):
        arch = "ARM"

    if "32" in processor_info:
        bit = "32"
    elif "64" in processor_info:
        bit = "64"

    if (arch is None) or (bit is None):
        raise EnvironmentError("Host info could not be parsed: {0}".format(str(host_info)))

    return Host(os=os, arch=arch, bit=bit)


####
# This is a mixing to do timing on a function. Use it like this:
#
# @timing
# def function_i_want_to_time():
#   ...
####
def timing(f):
    def wrap(*args):
        time1 = time.time()
        ret = f(*args)
        time2 = time.time()
        logging.debug('{:s} function took {:.3f} ms'.format(f.__name__, (time2 - time1) * 1000.0))
        return ret
    return wrap


####
# Run a shell commmand and print the output to the log file
# This is the public function that should be used to run commands from the shell in python environment
# @param cmd - command being run, either quoted or not quoted
# @param parameters - parameters string taken as is
# @param capture - boolean to determine if caller wants the output captured in any format.
# @param workingdir - path to set to the working directory before running the command.
# @param outfile - capture output to file of given path.
# @param outstream - capture output to a stream.
# @param environ - shell environment variables dictionary that replaces the one inherited from the
#                  current process.
# @param logging_level - log level to log output at.  Default is INFO
# @param raise_exception_on_nonzero - Setting to true causes exception to be raised if the cmd
#                                     return code is not zero.
#
# @return returncode of called cmd
####
def RunCmd(cmd, parameters, capture=True, workingdir=None, outfile=None, outstream=None, environ=None,
           logging_level=logging.INFO, raise_exception_on_nonzero=False):
    cmd = cmd.strip('"\'')
    if " " in cmd:
        cmd = '"' + cmd + '"'
    if parameters is not None:
        parameters = parameters.strip()
        cmd += " " + parameters
    starttime = datetime.datetime.now()
    logging.log(logging_level, "Cmd to run is: " + cmd)
    logging.log(logging_level, "------------------------------------------------")
    logging.log(logging_level, "--------------Cmd Output Starting---------------")
    logging.log(logging_level, "------------------------------------------------")
    c = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, cwd=workingdir, shell=True, env=environ)
    if(capture):
        outr = PropagatingThread(target=reader, args=(outfile, outstream, c.stdout, logging_level))
        outr.start()
        c.wait()
        outr.join()
    else:
        c.wait()

    endtime = datetime.datetime.now()
    delta = endtime - starttime
    endtime_str = "{0[0]:02}:{0[1]:02}".format(divmod(delta.seconds, 60))
    logging.log(logging_level, "------------------------------------------------")
    logging.log(logging_level, "--------------Cmd Output Finished---------------")
    logging.log(logging_level, "--------- Running Time (mm:ss): " + endtime_str + " ----------")
    logging.log(logging_level, "------------------------------------------------")

    if raise_exception_on_nonzero and c.returncode != 0:
        raise Exception("{0} failed with error code: {1}".format(cmd, c.returncode))
    return c.returncode

####
# Run a python script and print the output to the log file
# This is the public function that should be used to execute python scripts from the shell in python environment.
# The python script will be located using the path as if it was an executable.
#
# @param cmd - cmd string to run including parameters
# @param capture - boolean to determine if caller wants the output captured in any format.
# @param workingdir - path to set to the working directory before running the command.
# @param outfile - capture output to file of given path.
# @param outstream - capture output to a stream.
# @param environ - shell environment variables dictionary that replaces the one inherited from the
#                  current process.
#
# @return returncode of called cmd
####


def RunPythonScript(pythonfile, params, capture=True, workingdir=None, outfile=None, outstream=None,
                    environ=None, logging_level=logging.INFO, raise_exception_on_nonzero=False):
    # locate python file on path
    pythonfile.strip('"\'')
    if " " in pythonfile:
        pythonfile = '"' + pythonfile + '"'
    params.strip()
    logging.debug("RunPythonScript: {0} {1}".format(pythonfile, params))
    if(os.path.isabs(pythonfile)):
        logging.debug("Python Script was given as absolute path: %s" % pythonfile)
    elif(os.path.isfile(os.path.join(os.getcwd(), pythonfile))):
        pythonfile = os.path.join(os.getcwd(), pythonfile)
        logging.debug("Python Script was given as relative path: %s" % pythonfile)
    else:
        # loop thru path environment variable
        for a in os.getenv("PATH").split(os.pathsep):
            a = os.path.normpath(a)
            if os.path.isfile(os.path.join(a, pythonfile)):
                pythonfile = os.path.join(a, pythonfile)
                logging.debug("Python Script was found on the path: %s" % pythonfile)
                break
    params = pythonfile + " " + params
    return RunCmd(sys.executable, params, capture=capture, workingdir=workingdir, outfile=outfile,
                  outstream=outstream, environ=environ, logging_level=logging_level,
                  xraise_exception_on_nonzero=raise_exception_on_nonzero)
####
# Locally Sign input file using Windows SDK signtool.  This will use a local Pfx file.
# WARNING!!! : This should not be used for production signing as that process should follow stronger
#               security practices (HSM / smart cards / etc)
#
#  Signing is in format specified by UEFI authentacted variables
####


def DetachedSignWithSignTool(SignToolPath, ToSignFilePath, SignatureOutputFile, PfxFilePath,
                             PfxPass=None, Oid="1.2.840.113549.1.7.2", Eku=None):

    # check signtool path
    if not os.path.exists(SignToolPath):
        logging.error("Path to signtool invalid.  %s" % SignToolPath)
        return -1

    # Adjust for spaces in the path (when calling the command).
    if " " in SignToolPath:
        SignToolPath = '"' + SignToolPath + '"'

    OutputDir = os.path.dirname(SignatureOutputFile)
    # Signtool docs https://docs.microsoft.com/en-us/dotnet/framework/tools/signtool-exe
    # Signtool parameters from
    #   https://docs.microsoft.com/en-us/windows-hardware/manufacture/desktop/secure-boot-key-generation-and-signing-using-hsm--example  # noqa: E501
    # Search for "Secure Boot Key Generation and Signing Using HSM"
    params = 'sign /fd sha256 /p7ce DetachedSignedData /p7co ' + Oid + ' /p7 "' + \
             OutputDir + '" /f "' + PfxFilePath + '"'
    if Eku is not None:
        params += ' /u ' + Eku
    if PfxPass is not None:
        # add password if set
        params = params + ' /p ' + PfxPass
    params = params + ' /debug /v "' + ToSignFilePath + '" '
    ret = RunCmd(SignToolPath, params)
    if(ret != 0):
        logging.error("Signtool failed %d" % ret)
        return ret
    signedfile = os.path.join(OutputDir, os.path.basename(ToSignFilePath) + ".p7")
    if(not os.path.isfile(signedfile)):
        raise Exception("Output file doesn't eixst %s" % signedfile)

    shutil.move(signedfile, SignatureOutputFile)
    return ret

####
# Locally Sign input file using Windows SDK signtool.  This will use a local Pfx file.
# WARNING!!! : This should not be used for production signing as that process should follow
#              stronger security practices (HSM / smart cards / etc)
#
#  Signing is catalog format which is an attached signature
####


def CatalogSignWithSignTool(SignToolPath, ToSignFilePath, PfxFilePath, PfxPass=None):

    # check signtool path
    if not os.path.exists(SignToolPath):
        logging.error("Path to signtool invalid.  %s" % SignToolPath)
        return -1

    # Adjust for spaces in the path (when calling the command).
    if " " in SignToolPath:
        SignToolPath = '"' + SignToolPath + '"'

    OutputDir = os.path.dirname(ToSignFilePath)
    # Signtool docs https://docs.microsoft.com/en-us/dotnet/framework/tools/signtool-exe
    # todo: link to catalog signing documentation
    params = "sign /a /fd SHA256 /f " + PfxFilePath
    if PfxPass is not None:
        # add password if set
        params = params + ' /p ' + PfxPass
    params = params + ' /debug /v "' + ToSignFilePath + '" '
    ret = RunCmd(SignToolPath, params, workingdir=OutputDir)
    if(ret != 0):
        logging.error("Signtool failed %d" % ret)
    return ret


###
# Function to print a byte list as hex and optionally output ascii as well as
# offset within the buffer
###
def PrintByteList(ByteList, IncludeAscii=True, IncludeOffset=True, IncludeHexSep=True, OffsetStart=0):
    Ascii = ""
    for index in range(len(ByteList)):
        # Start of New Line
        if(index % 16 == 0):
            if(IncludeOffset):
                print("0x%04X -" % (index + OffsetStart), end='')

        # Midpoint of a Line
        if(index % 16 == 8):
            if(IncludeHexSep):
                print(" -", end='')

        # Print As Hex Byte
        print(" 0x%02X" % ByteList[index], end='')

        # Prepare to Print As Ascii
        if(ByteList[index] < 0x20) or (ByteList[index] > 0x7E):
            Ascii += "."
        else:
            Ascii += ("%c" % ByteList[index])

        # End of Line
        if(index % 16 == 15):
            if(IncludeAscii):
                print(" %s" % Ascii, end='')
            Ascii = ""
            print("")

    # Done - Lets check if we have partial
    if(index % 16 != 15):
        # Lets print any partial line of ascii
        if(IncludeAscii) and (Ascii != ""):
            # Pad out to the correct spot

            while(index % 16 != 15):
                print("     ", end='')
                if(index % 16 == 7):  # acount for the - symbol in the hex dump
                    if(IncludeOffset):
                        print("  ", end='')
                index += 1
            # print the ascii partial line
            print(" %s" % Ascii, end='')
            # print a single newline so that next print will be on new line
        print("")


# Simplified Comparison Function borrowed from StackOverflow...
# https://stackoverflow.com/questions/1714027/version-number-comparison
# With Python 3.0 help from:
# https://docs.python.org/3.0/whatsnew/3.0.html#ordering-comparisons
def version_compare(version1, version2):
    def normalize(v):
        return [int(x) for x in re.sub(r'(\.0+)*$', '', v).split(".")]
    (a, b) = (normalize(version1), normalize(version2))
    return (a > b) - (a < b)


def import_module_by_file_name(module_file_path):
    ''' Standard method of importing a Python file. Expecting absolute path. '''
    module_name = os.path.basename(module_file_path)
    spec = importlib.util.spec_from_file_location(module_name, module_file_path)

    if spec is None:
        raise RuntimeError(f"Expected module file named {module_file_path}")

    ImportedModule = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(ImportedModule)

    return ImportedModule


def locate_class_in_module(Module, DesiredClass):
    '''
    Given a module and a class, this function will return the subclass of DesiredClass in Module.
    Throws exception if two classes are found that fit the same criterea.
    '''

    DesiredClassInstance = None
    # Pull out the contents of the module that was provided
    module_contents = dir(Module)
    # Filter through the Module, we're only looking for classes.
    classList = [getattr(Module, obj) for obj in module_contents
                 if inspect.isclass(getattr(Module, obj))]
    for _class in classList:
        # Classes that the module import show up in this list too so we need
        # to make sure it's an INSTANCE of DesiredClass, not DesiredClass itself!
        if _class is not DesiredClass and issubclass(_class, DesiredClass):
            if DesiredClassInstance is not None:
                raise RuntimeError(f"Multiple instances were found:\n\t{DesiredClassInstance}\n\t{_class}")
            DesiredClassInstance = _class

    return DesiredClassInstance


if __name__ == '__main__':
    pass
    # Test code for printing a byte buffer
    # a = [0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d]
    # index = 0x55
    # while(index < 0x65):
    #     a.append(index)
    #     PrintByteList(a)
    #     index += 1



##
# Quick script to check that python code in the package
# aligns with pep8 and file encoding.  I have not found
# a way to enforce that with tools like flake8
#
# There must be a better way.  :)
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import glob
import os
import sys
import logging


def TestEncodingOk(apath, encodingValue):
    try:
        with open(apath, "rb") as fobj:
            fobj.read().decode(encodingValue)
    except Exception as exp:
        logging.critical("Encoding failure: file: {0} type: {1}".format(apath, encodingValue))
        logging.error("EXCEPTION: while processing {1} - {0}".format(exp, apath))
        return False
    return True


def TestFilenameLowercase(apath):
    if apath != apath.lower():
        logging.critical(f"Lowercase failure: file {apath} not lower case path")
        logging.error(f"\n\tLOWERCASE: {apath.lower()}\n\tINPUTPATH: {apath}")
        return False
    return True


def TestNoSpaces(apath):
    if " " in apath:
        logging.critical(f"NoSpaces failure: file {apath} has spaces in path")
        return False
    return True


def TestRequiredLicense(apath):
    lic = ["SPDX-License-Identifier: BSD-2-Clause-Patent"]
    try:
        with open(apath, "rb") as fobj:
            contents = fobj.read().decode()
            found = False
            for l in lic:
                if l in contents:
                    found = True
                    break
            if not found:
                logging.critical(f"License failure: file {apath} has incorrect, invalid, or unsupported license")
                return False
    except Exception as exp:
        logging.critical(f"License failure: Exception trying to read file: {apath}")
        logging.error("EXCEPTION: while processing {1} - {0}".format(exp, apath))
        return False
    return True


p = os.path.join(os.getcwd(), "edk2toollib")
pyfiles = glob.glob(os.path.join(p, "**", "*.py"), recursive=True)
error = 0
for a in pyfiles:
    aRelativePath = os.path.relpath(a, os.getcwd())
    if(not TestEncodingOk(a, "ascii")):
        error += 1
    if(not TestFilenameLowercase(aRelativePath)):
        error += 1
    if(not TestNoSpaces(aRelativePath)):
        error += 1
    if(not TestRequiredLicense(a)):
        error += 1

logging.critical(f"Found {error} error(s) in {len(pyfiles)} file(s)")
sys.exit(error)



# @file tpm2_policy_calc_test.py
# This file contains classes used to calculate TPM 2.0 policies
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import unittest
import edk2toollib.tpm.tpm2_policy_calc as t2pc


class TestPolicyLocality(unittest.TestCase):

    def test_create_with_empty_list(self):
        policy = t2pc.PolicyLocality(None)
        self.assertEqual(policy.get_bitfield(), 0)
        policy2 = t2pc.PolicyLocality(())
        self.assertEqual(policy2.get_bitfield(), 0)

    def test_create_with_base_localities(self):
        policy = t2pc.PolicyLocality([0, 2, 4])
        self.assertEqual(policy.get_bitfield(), 0b00010101)
        policy2 = t2pc.PolicyLocality([1, 2])
        self.assertEqual(policy2.get_bitfield(), 0b00000110)
        policy3 = t2pc.PolicyLocality([3])
        self.assertEqual(policy3.get_bitfield(), 0b00001000)
        policy4 = t2pc.PolicyLocality([57])
        self.assertEqual(policy4.get_bitfield(), 57)

    def test_create_with_invalid_localites(self):
        with self.assertRaises(ValueError):
            t2pc.PolicyLocality([5])
        with self.assertRaises(ValueError):
            t2pc.PolicyLocality([12])
        with self.assertRaises(ValueError):
            t2pc.PolicyLocality([31])
        with self.assertRaises(ValueError):
            t2pc.PolicyLocality([256])

    def test_create_with_mixed_lower_and_upper(self):
        with self.assertRaises(ValueError):
            t2pc.PolicyLocality([1, 4, 35])
        with self.assertRaises(ValueError):
            t2pc.PolicyLocality([36, 128])

    def test_get_buffer(self):
        self.assertEqual(t2pc.PolicyLocality([0, 2, 4]).get_buffer_for_digest(), bytearray.fromhex("0000016F" + "15"))
        self.assertEqual(t2pc.PolicyLocality([34]).get_buffer_for_digest(), bytearray.fromhex("0000016F" + "22"))


class TestPolicyCommandCode(unittest.TestCase):

    def test_create_with_no_code(self):
        with self.assertRaises(ValueError):
            t2pc.PolicyCommandCode(None)

    def test_create_with_invalid_code(self):
        with self.assertRaises(ValueError):
            t2pc.PolicyCommandCode("MonkeyValue")
        with self.assertRaises(ValueError):
            t2pc.PolicyCommandCode(12)
        with self.assertRaises(ValueError):
            t2pc.PolicyCommandCode({})

    def test_create_with_valid_codes(self):
        policy = t2pc.PolicyCommandCode('TPM_CC_Clear')
        self.assertEqual(policy.get_code(), 'TPM_CC_Clear')
        policy = t2pc.PolicyCommandCode('TPM_CC_ClearControl')
        self.assertEqual(policy.get_code(), 'TPM_CC_ClearControl')
        policy = t2pc.PolicyCommandCode('TPM_CC_Quote')
        self.assertEqual(policy.get_code(), 'TPM_CC_Quote')

    def test_get_buffer(self):
        self.assertEqual(t2pc.PolicyCommandCode('TPM_CC_Clear').get_buffer_for_digest(),
                         bytearray.fromhex("0000016C" + "00000126"))
        self.assertEqual(t2pc.PolicyCommandCode('TPM_CC_ClearControl').get_buffer_for_digest(),
                         bytearray.fromhex("0000016C" + "00000127"))


class TestPolicyTreeSolo(unittest.TestCase):

    def test_policy_command_code(self):
        expected_result_1 = bytearray.fromhex("940CFB4217BB1EDCF7FB41937CA974AA68E698AB78B8124B070113E211FD46FC")
        expected_result_2 = bytearray.fromhex("C4DFABCEDA8DE836C95661952892B1DEF7203AFB46FEFEC43FFCFC93BE540730")
        expected_result_3 = bytearray.fromhex("1D2DC485E177DDD0A40A344913CEEB420CAA093C42587D2E1B132B157CCB5DB0")

        test1 = t2pc.PolicyTreeSolo(t2pc.PolicyCommandCode("TPM_CC_ClearControl"))
        test2 = t2pc.PolicyTreeSolo(t2pc.PolicyCommandCode("TPM_CC_Clear"))
        test3 = t2pc.PolicyTreeSolo(t2pc.PolicyCommandCode("TPM_CC_NV_UndefineSpaceSpecial"))

        phash = t2pc.PolicyHasher('sha256')
        self.assertEqual(test1.get_policy(phash), expected_result_1)
        self.assertEqual(test2.get_policy(phash), expected_result_2)
        self.assertEqual(test3.get_policy(phash), expected_result_3)

    def test_policy_locality(self):
        expected_result = bytearray.fromhex("07039B45BAF2CC169B0D84AF7C53FD1622B033DF0A5DCDA66360AA99E54947CD")

        test = t2pc.PolicyTreeSolo(t2pc.PolicyLocality([3, 4]))

        phash = t2pc.PolicyHasher('sha256')
        self.assertEqual(test.get_policy(phash), expected_result)


class TestPolicyTreeAnd(unittest.TestCase):

    def test_single_and_should_match_solo(self):
        soloTest = t2pc.PolicyTreeSolo(t2pc.PolicyCommandCode("TPM_CC_Clear"))
        andTest = t2pc.PolicyTreeAnd([t2pc.PolicyCommandCode("TPM_CC_Clear")])

        phash = t2pc.PolicyHasher('sha256')
        self.assertEqual(soloTest.get_policy(phash), andTest.get_policy(phash))


class TestPolicyTreeOr(unittest.TestCase):

    def test_single_and_should_match_solo(self):
        expected_result = bytearray.fromhex("3F44FB41486D4A36A8ADCA2203E73A5068BFED5FDCE5092B9A3C6CCE8ABF3B0C")

        test1 = t2pc.PolicyTreeSolo(t2pc.PolicyCommandCode("TPM_CC_ClearControl"))
        test2 = t2pc.PolicyTreeSolo(t2pc.PolicyCommandCode("TPM_CC_Clear"))
        orTest = t2pc.PolicyTreeOr([test1, test2])

        phash = t2pc.PolicyHasher('sha256')
        self.assertEqual(orTest.get_policy(phash), expected_result)


class TestPolicyTree(unittest.TestCase):

    def test_complex_policy_1(self):
        expected_result = bytearray.fromhex("DFFDB6C8EAFCBE691E358882B18703121EAB40DE2386F7A8E7B4A06591E1F0EE")

        # Computation details:
        #   A = TPM2_PolicyLocality(3 & 4)
        #   B = TPM2_PolicyCommandCode(TPM_CC_NV_UndefineSpaceSpecial)
        #   C = TPM2_PolicyCommandCode(TPM_CC_NV_Write)
        #   policy = {{A} AND {C}} OR {{A} AND {B}}

        a = t2pc.PolicyLocality([3, 4])
        b = t2pc.PolicyCommandCode('TPM_CC_NV_UndefineSpaceSpecial')
        c = t2pc.PolicyCommandCode('TPM_CC_NV_Write')

        leg1 = t2pc.PolicyTreeAnd([a, c])
        leg2 = t2pc.PolicyTreeAnd([a, b])
        final = t2pc.PolicyTreeOr([leg1, leg2])

        phash = t2pc.PolicyHasher('sha256')
        self.assertEqual(final.get_policy(phash), expected_result)


if __name__ == '__main__':
    unittest.main()



# @file tpm2_policy_calc.py
# This file contains classes used to calculate TPM 2.0 policies
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##


import edk2toollib.tpm.tpm2_defs as t2d
import hashlib
import struct


# ========================================================================================
##
# POLICY TREE CLASSES
# These are used to describe a final policy structure.
# You can construct nodes to form complex policies from the policy primitive classes.
##
# PolicyTreeOr                                 <--- Tree Node
# /          \
# PolicyTreeSolo         PolicyTreeAnd                     <--- Tree Nodes
# /                     /        \
# PolicyCommandCode     PolicyLocality    PolicyCommandCode         <--- Primitives
##
# ========================================================================================


class PolicyHasher(object):
    def __init__(self, hash_type):
        if hash_type not in ['sha256', 'sha384']:
            raise ValueError("Invalid hash type '%s'!" % hash_type)

        self.hash_type = hash_type
        self.hash_size = {
            'sha256': 32,
            'sha384': 48
        }[hash_type]

    def get_size(self):
        return self.hash_size

    def hash(self, data):
        hash_obj = None
        if self.hash_type == 'sha256':
            hash_obj = hashlib.sha256()
        else:
            hash_obj = hashlib.sha384()

        hash_obj.update(data)

        return hash_obj.digest()


class PolicyCalculator(object):
    def __init__(self, primitive_dict, policy_tree):
        # For now, we'll leave this pretty sparse.
        # We should have WAY more testing for this stuff.
        self.primitive_dict = primitive_dict
        self.policy_tree = policy_tree

    def generate_digest(self, digest_type):
        pass


class PolicyTreeOr(object):
    def __init__(self, components):
        # OR connections can only be 8 digests long.
        # They CAN, however, be links of ORs.
        if len(components) > 8:
            raise ValueError("OR junctions cannot contain more than 8 sub-policies!")

        self.components = components

    def get_type(self):
        return 'or'

    def validate(self):
        result = True

        for component in self.components:
            # All components must be convertible into a policy.
            if not hasattr(component, 'get_policy'):
                result = False

            # All components must also be valid.
            if not hasattr(component, 'validate') or not component.validate():
                result = False

        return result

    def get_policy_buffer(self, hash_obj):
        concat_policy_buffer = b'\x00' * hash_obj.get_size()
        concat_policy_buffer += struct.pack(">L", t2d.TPM_CC_PolicyOR)
        concat_policy_buffer += b''.join([component.get_policy(hash_obj) for component in self.components])
        return concat_policy_buffer

    def get_policy(self, hash_obj):
        return hash_obj.hash(self.get_policy_buffer(hash_obj))


class PolicyTreeAnd(object):
    def __init__(self, components):
        # ANDs must only be composed of primitives. For simplicity, I guess.
        # Honestly, this has spiralled out of control, but something is better than nothing.
        for component in components:
            if not hasattr(component, 'get_buffer_for_digest'):
                raise ValueError("AND junctions must consist of primitives!")

        self.components = components

    def get_type(self):
        return 'and'

    def validate(self):
        return True

    def get_policy(self, hash_obj):
        current_digest = b'\x00' * hash_obj.get_size()
        for component in self.components:
            current_digest = hash_obj.hash(current_digest + component.get_buffer_for_digest())
        return current_digest


class PolicyTreeSolo(object):
    """This object should only be used to put a single policy claim under an OR"""

    def __init__(self, policy_obj):
        if not hasattr(policy_obj, 'get_buffer_for_digest'):
            raise ValueError("Supplied policy object is missing required functionality!")

        self.policy_obj = policy_obj

    def get_type(self):
        return 'solo'

    def validate(self):
        return True

    def get_policy_buffer(self, hash_obj):
        return (b'\x00' * hash_obj.get_size()) + self.policy_obj.get_buffer_for_digest()

    def get_policy(self, hash_obj):
        return hash_obj.hash(self.get_policy_buffer(hash_obj))


# ========================================================================================
##
# POLICY PRIMITIVES
# These classes are used to describe a single assertion (eg. PolicyLocality) and
# can be used with the PolicyTree classes to construct complex policies.
##
# ========================================================================================


class PolicyLocality(object):

    def __init__(self, localities):
        # Update the bitfield with the requested localities.
        if localities is not None:
            self.bitfield = self.calc_bitfield_from_list(localities)
        else:
            self.bitfield = 0b00000000

    def get_bitfield(self):
        return self.bitfield

    def calc_bitfield_from_list(self, localities):
        bitfield = 0b00000000

        # First, we need to validate all of the localities in the list.
        for value in localities:
            # If the value is in a bad range, we're done here.
            if not (0 <= value < 5) and not (32 <= value < 256):
                raise ValueError("Invalid locality '%d'!" % value)
            # An "upper" locality must be individual. Cannot combine with 0-4.
            if (32 <= value < 256) and len(localities) > 1:
                raise ValueError("Cannot combine locality '%d' with others!" % value)

        # If the list is empty... well, we're done.
        if len(localities) == 0:
            pass

        # Now, if we're an "upper" locality, that's a simple value.
        elif len(localities) == 1 and (32 <= localities[0] < 256):
            bitfield = localities[0]

        # We have to actually "think" to calculate the "lower" localities.
        else:
            for value in localities:
                bitfield |= 1 << value

        return bitfield

    def get_buffer_for_digest(self):
        # NOTE: We force big-endian to match the marshalling in the TPM.
        return struct.pack(">LB", t2d.TPM_CC_PolicyLocality, self.bitfield)


class PolicyCommandCode(object):

    def __init__(self, command_code_string=None):
        # Check to make sure that a command_code can be found.
        str_command_code_string = str(command_code_string)
        command_code = t2d.CommandCode.get_code(str_command_code_string)
        if command_code is None:
            raise ValueError("Command code '%s' unknown!" % str_command_code_string)
        self.command_code_string = str_command_code_string

    def get_code(self):
        return self.command_code_string

    def get_buffer_for_digest(self):
        # NOTE: We force big-endian to match the marshalling in the TPM.
        return struct.pack(">LL", t2d.CommandCode.get_code('TPM_CC_PolicyCommandCode'),
                           t2d.CommandCode.get_code(self.command_code_string))



# @file variable_format_Test.py
# Unit test harness for the VariableFormat module/classes.
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##


import unittest
import edk2toollib.uefi.edk2.variable_format as VF


class TestVariableHeader(unittest.TestCase):

    def test_set_name(self):
        var = VF.VariableHeader()

        test_name = "MyNewName"
        var.set_name(test_name)

        self.assertEqual(var.Name, test_name)

    def test_get_packed_name(self):
        var = VF.VariableHeader()

        test_name = "MyNewName"
        var.set_name(test_name)

        test_name_packed = bytes.fromhex('4D0079004E00650077004E0061006D0065000000')
        self.assertEqual(var.get_packed_name(), test_name_packed)



##
# unittest for ansi_handler
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
import unittest
import logging
from edk2toollib.log.ansi_handler import ColoredFormatter
from edk2toollib.log.ansi_handler import ColoredStreamHandler

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


class AnsiHandlerTest(unittest.TestCase):

    # we are mainly looking for exception to be thrown

    record = logging.makeLogRecord({"name": "", "level": logging.CRITICAL, "levelno": logging.CRITICAL,
                                    "levelname": "CRITICAL", "path": "test_path", "lineno": 0,
                                    "msg": "Test message"})
    record2 = logging.makeLogRecord({"name": "", "level": logging.INFO, "levelno": logging.INFO,
                                     "levelname": "INFO", "path": "test_path", "lineno": 0,
                                     "msg": "Test message"})

    def test_colored_formatter_init(self):
        formatter = ColoredFormatter("%(levelname)s - %(message)s")
        # if we didn't throw an exception, then we are good
        self.assertNotEqual(formatter, None)

    def test_colored_formatter_to_output_ansi(self):
        formatter = ColoredFormatter("%(levelname)s - %(message)s")

        output = formatter.format(AnsiHandlerTest.record)
        self.assertNotEqual(output, None)
        CSI = '\033['
        self.assertGreater(len(output), 0, "We should have some output")
        self.assertFalse((CSI not in output), "There was supposed to be a ANSI control code in that %s" % output)

    def test_color_handler_to_strip_ansi(self):
        stream = StringIO()
        # make sure we set out handler to strip the control sequence
        handler = ColoredStreamHandler(stream, strip=True, convert=False)
        formatter = ColoredFormatter("%(levelname)s - %(message)s")
        handler.formatter = formatter
        handler.level = logging.NOTSET

        handler.emit(AnsiHandlerTest.record)
        handler.flush()

        CSI = '\033['

        # check for ANSI escape code in stream
        stream.seek(0)
        lines = stream.readlines()
        self.assertGreater(len(lines), 0, "We should have some output %s" % lines)
        for line in lines:
            if CSI in line:
                self.fail("A control sequence was not stripped! %s" % lines)

    def test_color_handler_not_strip_ansi(self):
        stream = StringIO()
        formatter = ColoredFormatter("%(levelname)s - %(message)s")
        handler = ColoredStreamHandler(stream, strip=False, convert=False)
        handler.formatter = formatter
        handler.level = logging.NOTSET

        handler.emit(AnsiHandlerTest.record2)
        handler.flush()

        CSI = '\033['

        found_csi = False
        stream.seek(0)
        lines = stream.readlines()
        self.assertGreater(len(lines), 0, "We should have some output %s" % lines)
        for line in lines:
            if CSI in line:
                found_csi = True
        self.assertTrue(found_csi, "We are supposed to to have found an ANSI control character %s" % lines)



## @file
# UnitTest for cat_generator.py based on various architecture/OS
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import os
import unittest
from edk2toollib.windows.capsule.cat_generator import CatGenerator


class CatGeneratorTest(unittest.TestCase):

    def test_win10_OS(self):
        o = CatGenerator("x64", "win10")
        self.assertEqual(o.OperatingSystem, "10")

    def test_10_OS(self):
        o = CatGenerator("x64", "10")
        self.assertEqual(o.OperatingSystem, "10")

    def test_10_AU_OS(self):
        o = CatGenerator("x64", "10_AU")
        self.assertEqual(o.OperatingSystem, "10_AU")

    def test_10_RS2_OS(self):
        o = CatGenerator("x64", "10_RS2")
        self.assertEqual(o.OperatingSystem, "10_RS2")

    def test_10_RS3_OS(self):
        o = CatGenerator("x64", "10_RS3")
        self.assertEqual(o.OperatingSystem, "10_RS3")

    def test_10_RS4_OS(self):
        o = CatGenerator("x64", "10_RS4")
        self.assertEqual(o.OperatingSystem, "10_RS4")

    def test_win10Server_OS(self):
        o = CatGenerator("x64", "Server10")
        self.assertEqual(o.OperatingSystem, "Server10")

    def test_Server2016_OS(self):
        o = CatGenerator("x64", "Server2016")
        self.assertEqual(o.OperatingSystem, "Server2016")

    def test_ServerRS2_OS(self):
        o = CatGenerator("x64", "ServerRS2")
        self.assertEqual(o.OperatingSystem, "ServerRS2")

    def test_ServerRS3_OS(self):
        o = CatGenerator("x64", "ServerRS3")
        self.assertEqual(o.OperatingSystem, "ServerRS3")

    def test_ServerRS4_OS(self):
        o = CatGenerator("x64", "ServerRS4")
        self.assertEqual(o.OperatingSystem, "ServerRS4")

    def test_invalid_OS(self):
        with self.assertRaises(ValueError):
            CatGenerator("x64", "Invalid Junk")

    def test_x64_arch(self):
        o = CatGenerator("x64", "win10")
        self.assertEqual(o.Arch, "X64")

    def test_amd64_arch(self):
        o = CatGenerator("amd64", "win10")
        self.assertEqual(o.Arch, "X64")

    def test_arm_arch(self):
        o = CatGenerator("arm", "win10")
        self.assertEqual(o.Arch, "ARM")

    def test_arm64_arch(self):
        o = CatGenerator("arm64", "win10")
        self.assertEqual(o.Arch, "ARM64")

    def test_aarch64_arch(self):
        o = CatGenerator("aarch64", "win10")
        self.assertEqual(o.Arch, "ARM64")

    def test_invalid_arch(self):
        with self.assertRaises(ValueError):
            CatGenerator("Invalid Arch", "win10")

    def test_invalid_pathtotool(self):
        o = CatGenerator("amd64", "10")
        with self.assertRaises(Exception) as cm:
            o.MakeCat("garbage", os.path.join("c:", "test", "badpath", "inf2cat.exe"))
        self.assertTrue(str(cm.exception).startswith("Can't find Inf2Cat on this machine."))



## @file override_parser.py
# Contains classes to help with the parsing of INF files that
# may contain OVERRIDE information.
#
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

import os
import datetime

FORMAT_VERSION_1 = (1, 4)   # Version 1: #OVERRIDE : VERSION | PATH_TO_MODULE | HASH | YYYY-MM-DDThh-mm-ss


class OpParseError(Exception):
    PE_VER = 'VERSION'
    PE_PATH = 'PATH'
    PE_HASH = 'HASH'
    PE_DATE = 'DATE'

    def __init__(self, my_type):
        if my_type not in (OpParseError.PE_VER, OpParseError.PE_PATH,
                           OpParseError.PE_HASH, OpParseError.PE_DATE):
            raise ValueError("Unknown type '%s'" % my_type)
        self.type = my_type

    def __str__(self):
        return repr(self.type)


class OverrideParser(object):
    """
    OverrideParser is a simple file parser for .inf files that
    contain OVERRIDE data (i.e. overriding other .infs).
    Creating the object can be done by passing either a valid file path
    or a string containing the contents of an .inf file.

    Will raise an exception if the file doesn't exist or if the contents
    do not contain any OVERRIDE data.

    NOTE: There is an argument to be made that this class should actually be
          a subclass of InfParser, however, the InfParser is looking for far
          more details and has a much higher overhead. During a parser refactor,
          this should be considered.

    ALSO NOTE: There is a pattern used here where the object parses during
               instantiation. This pattern does not necessarily match the other
               parsers. The pros and cons of this should also be weighed during
               any parser refactor.
    """
    def __init__(self, file_path=None, inf_contents=None):
        super(OverrideParser, self).__init__()

        # Make sure that at least some data is provided.
        if file_path is None and inf_contents is None:
            raise ValueError("file_path or inf_contents is required.")
        # Make sure not too much data is provided.
        if file_path is not None and inf_contents is not None:
            raise ValueError("Only provide file_path or inf_contents. (%s, %s)" % (file_path, inf_contents))
        # If a file path was provided, make sure it exists.
        if file_path is not None:
            if not os.path.isfile(file_path):
                raise ValueError("File path '%s' does not exist." % file_path)

        self.file_path = os.path.abspath(file_path) if file_path is not None else 'String Buffer'

        # Set up the contents for parsing.
        parse_contents = inf_contents
        if file_path is not None:
            with open(file_path, 'r') as file:
                parse_contents = file.read()
        if not parse_contents:
            raise ValueError("Failed to read contents of file '%s'." % self.file_path)

        self.override_lines = self.get_override_lines(parse_contents)

        # If no override lines were found, we're basically done here.
        if not self.override_lines:
            raise ValueError("File '%s' did not contain any override lines." % self.file_path)

        self.overrides = []
        for override_line in self.override_lines:
            try:
                self.overrides.append(self.parse_override_line(override_line['line']))
            except OpParseError as pe:
                raise ValueError("Parse error '%s' occurred while processing line %d of '%s'." %
                                 (pe, override_line['lineno'], override_line['line']))

    @staticmethod
    def get_override_lines(parse_contents):
        parse_lines = parse_contents.split('\n')
        result = []

        for i in range(0, len(parse_lines)):
            if parse_lines[i].strip().upper().startswith("#OVERRIDE"):
                result.append({'lineno': i + 1, 'line': parse_lines[i].strip()})

        return result

    @staticmethod
    def parse_override_line(line_contents):
        result = {}

        # Split the override string into pieces.
        # First the #OVERRIDE, which is separated by a :.
        # Then everything else by |.
        line_parts = line_contents.split(":")
        line_parts = [part.strip() for part in line_parts[1].split("|")]

        # Step 1: Check version and number of blocks in this entry
        try:
            result['version'] = int(line_parts[0])
        except ValueError:
            raise OpParseError(OpParseError.PE_VER)

        # Verify this is a known version and has valid number of entries
        if not ((result['version'] == FORMAT_VERSION_1[0]) and (len(line_parts) == FORMAT_VERSION_1[1])):
            raise OpParseError(OpParseError.PE_VER)

        # Step 2: Process the path to overridden module
        # Normalize the path to support different slashes.
        result['original_path'] = os.path.normpath(line_parts[1])

        # Step 3: Grep hash entry
        result['current_hash'] = line_parts[2]

        # Step 4: Parse the time of hash generation
        try:
            result['datetime'] = datetime.datetime.strptime(line_parts[3], "%Y-%m-%dT%H-%M-%S")
        except ValueError:
            raise OpParseError(OpParseError.PE_DATE)

        return result



# @file wincert.py
# Code to work with UEFI WinCert data
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##


import struct
import uuid
from edk2toollib.utility_functions import PrintByteList


class WinCertPkcs1(object):
    STATIC_STRUCT_SIZE = (4 + 2 + 2 + 16)
    EFI_HASH_SHA256 = uuid.UUID("{51AA59DE-FDF2-4EA3-BC63-875FB7842EE9}")  # EFI_HASH_SHA256 guid defined by UEFI spec

    def __init__(self, filestream=None):
        if(filestream is None):
            self.Hdr_dwLength = WinCertPkcs1.STATIC_STRUCT_SIZE
            self.Hdr_wRevision = WinCert.REVISION
            self.Hdr_wCertificateType = WinCert.WIN_CERT_TYPE_EFI_PKCS115
            self.HashAlgorithm = None
            self.CertData = None
        else:
            self.PopulateFromFileStream(filestream)

    def AddCertData(self, fs):
        if(self.CertData is not None):
            raise Exception("Cert Data not 0")
        if(self.HashAlgorithm is None):
            raise Exception("You must set the Hash Algorithm first")
        self.CertData = fs.read()
        self.Hdr_dwLength = self.Hdr_dwLength + len(self.CertData)
    #
    # Method to un-serialize from a filestream
    #

    def PopulateFromFileStream(self, fs):
        if(fs is None):
            raise Exception("Invalid File stream")

        # only populate from file stream those parts that are complete in the file stream
        offset = fs.tell()
        fs.seek(0, 2)
        end = fs.tell()
        fs.seek(offset)

        if((end - offset) < WinCertPkcs1.STATIC_STRUCT_SIZE):  # size of the static header data
            raise Exception("Invalid file stream size")

        self.Hdr_dwLength = struct.unpack("=I", fs.read(4))[0]
        self.Hdr_wRevision = struct.unpack("=H", fs.read(2))[0]
        self.Hdr_wCertificateType = struct.unpack("=H", fs.read(2))[0]
        self.HashAlgorithm = uuid.UUID(bytes_le=fs.read(16))
        self.CertData = None

        if((end - fs.tell()) < 1):
            raise Exception("Invalid File stream. No data for signature cert data")

        if((end - fs.tell()) < (self.Hdr_dwLength - WinCertPkcs1.STATIC_STRUCT_SIZE)):
            raise Exception("Invalid file stream size")

        self.CertData = memoryview(fs.read(self.Hdr_dwLength - WinCertPkcs1.STATIC_STRUCT_SIZE))

    def Print(self):
        print("WinCertPKCS115")
        print("  Hdr_dwLength:         0x%X" % self.Hdr_dwLength)
        print("  Hdr_wRevision:        0x%X" % self.Hdr_wRevision)
        print("  Hdr_wCertificateType: 0x%X" % self.Hdr_wCertificateType)
        print("  Hash Guid:            %s" % str(self.HashAlgorithm))
        print("  CertData:             ")
        cdl = self.CertData.tolist()
        PrintByteList(cdl)

    def Write(self, fs):
        fs.write(struct.pack("=I", self.Hdr_dwLength))
        fs.write(struct.pack("=H", self.Hdr_wRevision))
        fs.write(struct.pack("=H", self.Hdr_wCertificateType))
        fs.write(self.HashAlgorithm.bytes_le)
        fs.write(self.CertData)

##
# WIN_CERT_UEFI_GUID
##


class WinCertUefiGuid(object):
    STATIC_STRUCT_SIZE = (4 + 2 + 2 + 16)
    PKCS7Guid = uuid.UUID("{4aafd29d-68df-49ee-8aa9-347d375665a7}")  # PKCS7 guid defined by UEFI spec

    def __init__(self, filestream=None):
        if(filestream is None):
            self.Hdr_dwLength = WinCertUefiGuid.STATIC_STRUCT_SIZE
            self.Hdr_wRevision = WinCert.REVISION
            self.Hdr_wCertificateType = WinCert.WIN_CERT_TYPE_EFI_GUID
            self.CertType = WinCertUefiGuid.PKCS7Guid
            self.CertData = None
        else:
            self.PopulateFromFileStream(filestream)

    def AddCertData(self, fs):
        if(self.CertData is not None):
            raise Exception("Cert Data not 0")
        self.CertData = memoryview(fs.read())
        self.Hdr_dwLength = self.Hdr_dwLength + len(self.CertData)
    #
    # Method to un-serialize from a filestream
    #

    def PopulateFromFileStream(self, fs):
        if(fs is None):
            raise Exception("Invalid File stream")

        # only populate from file stream those parts that are complete in the file stream
        offset = fs.tell()
        fs.seek(0, 2)
        end = fs.tell()
        fs.seek(offset)

        if((end - offset) < WinCertUefiGuid.STATIC_STRUCT_SIZE):  # size of the static header data
            raise Exception("Invalid file stream size")

        self.Hdr_dwLength = struct.unpack("=I", fs.read(4))[0]
        self.Hdr_wRevision = struct.unpack("=H", fs.read(2))[0]
        self.Hdr_wCertificateType = struct.unpack("=H", fs.read(2))[0]
        self.CertType = uuid.UUID(bytes_le=fs.read(16))
        self.CertData = None

        if((end - fs.tell()) < 1):
            raise Exception("Invalid File stream. No data for signature cert data")

        if((end - fs.tell()) < (self.Hdr_dwLength - WinCertUefiGuid.STATIC_STRUCT_SIZE)):
            raise Exception("Invalid file stream size ")

        self.CertData = memoryview(fs.read(self.Hdr_dwLength - WinCertUefiGuid.STATIC_STRUCT_SIZE))

    def Print(self):
        print("WinCertUefiGuid")
        print("  Hdr_dwLength:         0x%X" % self.Hdr_dwLength)
        print("  Hdr_wRevision:        0x%X" % self.Hdr_wRevision)
        print("  Hdr_wCertificateType: 0x%X" % self.Hdr_wCertificateType)
        print("  CertType:             %s" % str(self.CertType))
        print("  CertData:             ")
        cdl = self.CertData.tolist()
        PrintByteList(cdl)

    def Write(self, fs):
        fs.write(struct.pack("=I", self.Hdr_dwLength))
        fs.write(struct.pack("=H", self.Hdr_wRevision))
        fs.write(struct.pack("=H", self.Hdr_wCertificateType))
        fs.write(self.CertType.bytes_le)
        fs.write(self.CertData)


class WinCert(object):
    STATIC_STRUCT_SIZE = 8
    # WIN_CERTIFICATE.wCertificateTypes UEFI Spec defined
    WIN_CERT_TYPE_NONE = 0x0000
    WIN_CERT_TYPE_PKCS_SIGNED_DATA = 0x0002
    WIN_CERT_TYPE_EFI_PKCS115 = 0x0EF0
    WIN_CERT_TYPE_EFI_GUID = 0x0EF1
    # Revision
    REVISION = 0x200

    #
    # this method is a factory
    #
    @staticmethod
    def Factory(fs):
        if(fs is None):
            raise Exception("Invalid File stream")

        # only populate from file stream those parts that are complete in the file stream
        offset = fs.tell()
        fs.seek(0, 2)
        end = fs.tell()
        fs.seek(offset)

        if((end - offset) < WinCert.STATIC_STRUCT_SIZE):  # size of the static header data
            raise Exception("Invalid file stream size")
        # 1 read len
        # 2 read revision
        # 3 read cert type
        fs.seek(4, 1)  # seeking past Hdr_dwLength
        fs.seek(2, 1)  # seeking past Hdr_wRevision
        Hdr_wCertificateType = struct.unpack("=H", fs.read(2))[0]

        fs.seek(offset)

        if(Hdr_wCertificateType == WinCert.WIN_CERT_TYPE_EFI_GUID):
            return WinCertUefiGuid(fs)
        elif(Hdr_wCertificateType == WinCert.WIN_CERT_TYPE_EFI_PKCS115):
            return WinCertPkcs1(fs)
        else:
            return None
